# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Functionality for brax bodies."""

from flax import struct
import jax
import jax.numpy as jnp

from brax.physics import config_pb2
from brax.physics import math
from brax.physics.base import P, QP, euler_to_quat, vec_to_np

@struct.dataclass
class Body(object):
  """A body is a solid, non-deformable object with some mass and shape.

  Attributes:
    idx: Index of where body is found in the system.
    inertia: (3, 3) Inverse Inertia matrix represented in body frame.
    mass: Mass of the body.
    active: whether the body is effected by physics calculations
  """
  idx: jnp.ndarray
  inertia: jnp.ndarray
  mass: jnp.ndarray
  active: jnp.ndarray

  @classmethod
  def from_config(cls, config: config_pb2.Config) -> 'Body':
    """Returns Body from a brax config."""
    bodies = []
    for idx, body in enumerate(config.bodies):
      frozen = jnp.sum(
          vec_to_np(body.frozen.position) + vec_to_np(body.frozen.rotation))
      bodies.append(
          cls(
              idx=jnp.array(idx),
              inertia=jnp.linalg.inv(jnp.diag(vec_to_np(body.inertia))),
              mass=jnp.array(body.mass),
              active=jnp.array(jnp.sum(frozen) != 6),
          ))
    return jax.tree_multimap((lambda *args: jnp.stack(args)), *bodies)

  def impulse(self, qp: QP, impulse: jnp.ndarray, pos: jnp.ndarray) -> P:
    """Calculates updates to state information based on an impulse.

    Args:
      qp: State data of the system
      impulse: Impulse vector
      pos: Location of the impulse relative to the body's center of mass

    Returns:
      dP: An impulse to apply to this body
    """
    dvel = impulse / self.mass
    dang = jnp.matmul(self.inertia, jnp.cross(pos - qp.pos, impulse))
    return P(vel=dvel, ang=dang)

def min_z(qp: QP, body: config_pb2.Body) -> float:
  """Returns the lowest z of all the colliders in a body."""
  result = float('inf')

  for col in body.colliders:
    if col.HasField('sphere'):
      sphere_pos = math.rotate(vec_to_np(col.position), qp.rot)
      min_z = qp.pos[2] + sphere_pos[2] - col.sphere.radius
      result = jnp.min(jnp.array([result, min_z]))
    elif col.HasField('capsule'):
      axis = math.rotate(jnp.array([0., 0., 1.]), euler_to_quat(col.rotation))
      length = col.capsule.length / 2 - col.capsule.radius
      for end in (-1, 1):
        sphere_pos = vec_to_np(col.position) + end * axis * length
        sphere_pos = math.rotate(sphere_pos, qp.rot)
        min_z = qp.pos[2] + sphere_pos[2] - col.capsule.radius
        result = jnp.min(jnp.array([result, min_z]))
    elif col.HasField('box'):
      corners = [(i % 2 * 2 - 1, 2 * (i // 4) - 1, i // 2 % 2 * 2 - 1)
                 for i in range(8)]
      corners = jnp.array(corners, dtype=jnp.float32)
      for corner in corners:
        corner = corner * vec_to_np(col.box.halfsize)
        corner = math.rotate(corner, euler_to_quat(col.rotation))
        corner = corner + vec_to_np(col.position)
        corner = math.rotate(corner, qp.rot) + qp.pos
        result = jnp.min(jnp.array([result, corner[2]]))
    else:
      # ignore planes and other stuff
      result = jnp.min(jnp.array([result, 0.0]))

  return result


# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Augmented Random Search training.

See: https://arxiv.org/pdf/1803.07055.pdf
"""

import time
from typing import Any, Callable, Dict, Optional

from absl import logging
from brax import envs
from brax.training import env
from brax.training import networks
from brax.training import normalization
import flax
import jax
import jax.numpy as jnp
import optax

Params = Any


@flax.struct.dataclass
class TrainingState:
  """Contains training state for the learner."""
  key: jnp.ndarray
  normalizer_params: Params
  policy_params: Params


def make_ars_model(act_size: int, obs_size: int):
  return networks.FeedForwardModel(
      init=lambda _: jnp.zeros((obs_size, act_size)),
      apply=lambda m, o: jnp.matmul(o, m))


def get_policy_head(head_type):
  def head(params):
    if not head_type:
      return params
    if head_type == 'clip':
      return jnp.clip(params, -1, 1)
    if head_type == 'tanh':
      return jnp.tanh(params)
    assert f'policy head type {head_type} is not known'
  return head


def train(
    environment_fn: Callable[..., envs.Env],
    num_timesteps: int = 100,
    log_frequency: int = 1,
    episode_length: int = 1000,
    action_repeat: int = 1,
    num_eval_envs: int = 128,
    seed: int = 0,
    normalize_observations: bool = False,
    step_size: float = 0.015,
    max_devices_per_host: Optional[int] = None,
    number_of_directions: int = 60,
    exploration_noise_std: float = 0.025,
    top_directions: int = 20,
    head_type: str = '',
    reward_shift: float = 0.0,
    progress_fn: Optional[Callable[[int, Dict[str, Any]], None]] = None,
):
  """ARS."""
  # TODO: pmap it
  max_devices_per_host = 1

  xt = time.time()
  top_directions = min(top_directions, number_of_directions)
  num_envs = number_of_directions * 2  # antitethic

  process_count = jax.process_count()
  process_id = jax.process_index()
  local_device_count = jax.local_device_count()
  local_devices_to_use = local_device_count
  if max_devices_per_host:
    local_devices_to_use = min(local_devices_to_use, max_devices_per_host)
  logging.info(
      'Device count: %d, process count: %d (id %d), local device count: %d, '
      'devices to be used count: %d',
      jax.device_count(), process_count, process_id, local_device_count,
      local_devices_to_use)

  key = jax.random.PRNGKey(seed)
  key, key_model, key_env, key_eval = jax.random.split(key, 4)

  core_env = environment_fn(
      action_repeat=action_repeat,
      batch_size=num_envs // local_devices_to_use // process_count,
      episode_length=episode_length)
  first_state = core_env.reset(key_env)
  step_fn = core_env.step

  core_eval_env = environment_fn(
      action_repeat=action_repeat,
      batch_size=num_eval_envs,
      episode_length=episode_length)
  eval_first_state, eval_step_fn = env.wrap_for_eval(core_eval_env, key_eval)

  _, obs_size = eval_first_state.core.obs.shape

  policy_model = make_ars_model(core_env.action_size, obs_size)
  policy_head = get_policy_head(head_type)

  normalizer_params, obs_normalizer_update_fn, obs_normalizer_apply_fn = (
      normalization.create_observation_normalizer(
          obs_size, normalize_observations, num_leading_batch_dims=1,
          apply_clipping=False))

  policy_params = policy_model.init(key_model)

  def do_one_step_eval(carry, unused_target_t):
    state, policy_params, normalizer_params = carry
    obs = obs_normalizer_apply_fn(normalizer_params, state.core.obs)
    actions = policy_head(policy_model.apply(policy_params, obs))
    nstate = eval_step_fn(state, actions)
    return (nstate, policy_params, normalizer_params), ()

  @jax.jit
  def run_eval(state, policy_params, normalizer_params) -> env.EvalEnvState:
    (state, _, _), _ = jax.lax.scan(
        do_one_step_eval, (state, policy_params, normalizer_params), (),
        length=episode_length // action_repeat)
    return state

  @jax.vmap
  def training_inference(params, obs):
    return policy_model.apply(params, obs)

  def do_one_step(args):
    state, policy_params, cumulative_reward, active_episode, normalizer_params, new_normalizer_params = args
    obs = obs_normalizer_apply_fn(normalizer_params, state.obs)
    actions = policy_head(training_inference(policy_params, obs))
    nstate = step_fn(state, actions)
    cumulative_reward = cumulative_reward + (nstate.reward -
                                             reward_shift) * active_episode
    new_active_episode = active_episode * (1 - nstate.done)
    new_normalizer_params = obs_normalizer_update_fn(new_normalizer_params,
                                                     state.obs,
                                                     active_episode)
    return nstate, policy_params, cumulative_reward, new_active_episode, normalizer_params, new_normalizer_params

  def run_ars_eval(state, params, normalizer_params):
    cumulative_reward = jnp.zeros(state.obs.shape[0])
    active_episode = jnp.ones(state.obs.shape[0])
    _, _, cumulative_reward, _, _, normalizer_params = jax.lax.while_loop(
        lambda a: jnp.sum(a[3]) > 0, do_one_step,
        (state, params, cumulative_reward, active_episode, normalizer_params,
         normalizer_params))
    return cumulative_reward, normalizer_params

  def add_noise(params, key):
    noise = jax.random.normal(key, shape=params.shape, dtype=params.dtype)
    params_with_noise = params + noise * exploration_noise_std
    anit_params_with_noise = params - noise * exploration_noise_std
    return params_with_noise, anit_params_with_noise, noise

  @jax.jit
  def ars_one_epoch(state, training_state):
    params = jnp.repeat(jnp.expand_dims(training_state.policy_params, axis=0),
                        num_envs // 2, axis=0)

    key, key_petr = jax.random.split(training_state.key)
    # generate perturbations
    params_with_noise, params_with_anti_noise, noise = add_noise(
        params, key_petr)

    pparams = jnp.concatenate([params_with_noise, params_with_anti_noise],
                              axis=0)

    eval_scores, normalizer_params = run_ars_eval(
        state, pparams, training_state.normalizer_params)

    reward_plus, reward_minus = jnp.split(eval_scores, 2, axis=0)
    reward_max = jnp.maximum(reward_plus, reward_minus)
    reward_rank = jnp.argsort(jnp.argsort(-reward_max))
    reward_weight = jnp.where(reward_rank < top_directions, 1, 0)
    reward_weight_double = jnp.concatenate([reward_weight, reward_weight],
                                           axis=0)
    reward_std = jnp.std(eval_scores, where=reward_weight_double)

    noise = jnp.sum(jnp.transpose(jnp.transpose(noise) * reward_weight *
                                  (reward_plus - reward_minus)), axis=0)

    policy_params = (training_state.policy_params +
                     step_size / (top_directions * reward_std) * noise)

    metrics = {
        'params_norm': optax.global_norm(policy_params),
        'eval_scores_mean': jnp.mean(eval_scores),
        'eval_scores_std': jnp.std(eval_scores),
        'reward_std': reward_std,
        'weights': jnp.mean(reward_weight),
    }
    return TrainingState(
        key=key,
        normalizer_params=normalizer_params,
        policy_params=policy_params), metrics

  training_state = TrainingState(key=key,
                                 normalizer_params=normalizer_params,
                                 policy_params=policy_params)

  training_walltime = 0
  eval_walltime = 0
  sps = 0
  eval_sps = 0
  metrics = {}
  summary = {}
  state = first_state
  it = -1

  while True:
    it += 1
    logging.info('starting iteration %s %s', it, time.time() - xt)
    t = time.time()
    num_process_env_steps = int(
        training_state.normalizer_params[0]) * action_repeat

    if process_id == 0 and it % log_frequency == 0:
      eval_state = run_eval(eval_first_state,
                            training_state.policy_params,
                            training_state.normalizer_params)
      eval_state.completed_episodes.block_until_ready()
      eval_walltime += time.time() - t
      eval_sps = (
          episode_length * eval_first_state.core.reward.shape[0] /
          (time.time() - t))
      avg_episode_length = (
          eval_state.completed_episodes_steps / eval_state.completed_episodes)
      metrics = dict(
          dict({
              f'eval/episode_{name}': value / eval_state.completed_episodes
              for name, value in eval_state.completed_episodes_metrics.items()
          }),
          **dict({
              f'train/{name}': value for name, value in summary.items()
          }),
          **dict({
              'eval/completed_episodes': eval_state.completed_episodes,
              'eval/episode_length': avg_episode_length,
              'train/completed_episodes': it * num_envs,
              'speed/sps': sps,
              'speed/eval_sps': eval_sps,
              'speed/training_walltime': training_walltime,
              'speed/eval_walltime': eval_walltime,
              'speed/timestamp': training_walltime,
          }))
      logging.info('Step %s metrics %s', num_process_env_steps, metrics)
      if progress_fn:
        progress_fn(num_process_env_steps, metrics)

    if num_process_env_steps > num_timesteps:
      break

    t = time.time()
    # optimization
    training_state, summary = ars_one_epoch(state, training_state)
    # Don't override state with new_state. For environments with variable
    # episode length we still want to start from a 'reset', not from where the
    # last run finished.

    jax.tree_map(lambda x: x.block_until_ready(), training_state)
    sps = (int(training_state.normalizer_params[0]) * action_repeat -
           num_process_env_steps) / (
               time.time() - t)
    training_walltime += time.time() - t

  _, inference = make_params_and_inference_fn(core_env.observation_size,
                                              core_env.action_size,
                                              normalize_observations,
                                              head_type)
  params = training_state.normalizer_params, training_state.policy_params

  return (inference, params, metrics)


def make_params_and_inference_fn(observation_size, action_size,
                                 normalize_observations, head_type=None):
  """Creates params and inference function for the ES agent."""
  obs_normalizer_params, obs_normalizer_apply_fn = normalization.make_data_and_apply_fn(
      observation_size, normalize_observations, apply_clipping=False)
  policy_head = get_policy_head(head_type)
  policy_model = make_ars_model(action_size, observation_size)

  def inference_fn(params, obs, unused_rng):
    normalizer_params, policy_params = params
    obs = obs_normalizer_apply_fn(normalizer_params, obs)
    action = policy_head(policy_model.apply(policy_params, obs))
    return action

  params = (obs_normalizer_params, policy_model.init(jax.random.PRNGKey(0)))
  return params, inference_fn



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Soft Actor-Critic training.

See: https://arxiv.org/pdf/1812.05905.pdf
"""

import time
from typing import Any, Callable, Dict, Mapping, Optional, Tuple

from absl import logging
from brax import envs
from brax.io import model
from brax.training import distribution
from brax.training import env
from brax.training import networks
from brax.training import normalization
from brax.training.types import Params
from brax.training.types import PRNGKey
import flax
from flax import linen
import jax
import jax.numpy as jnp
import numpy as onp
import optax


Metrics = Mapping[str, jnp.ndarray]


@flax.struct.dataclass
class Transition:
  """Contains data for one environment step."""
  o_tm1: jnp.ndarray
  a_tm1: jnp.ndarray
  r_t: jnp.ndarray
  o_t: jnp.ndarray
  d_t: jnp.ndarray  # discount (1-done)


# The rewarder allows to change the reward of before the learner trains.
RewarderState = Any
RewarderInit = Callable[[int, PRNGKey], RewarderState]
ComputeReward = Callable[[RewarderState, Transition, PRNGKey],
                         Tuple[RewarderState, jnp.ndarray, Metrics]]
Rewarder = Tuple[RewarderInit, ComputeReward]


@flax.struct.dataclass
class ReplayBuffer:
  """Contains data related to a replay buffer."""
  data: jnp.ndarray
  current_position: jnp.ndarray
  current_size: jnp.ndarray


@flax.struct.dataclass
class TrainingState:
  """Contains training state for the learner."""
  policy_optimizer_state: optax.OptState
  policy_params: Params
  q_optimizer_state: optax.OptState
  q_params: Params
  target_q_params: Params
  key: PRNGKey
  steps: jnp.ndarray
  alpha_optimizer_state: optax.OptState
  alpha_params: Params
  normalizer_params: Params
  # The is passed to the rewarder to update the reward.
  rewarder_state: Any


def make_sac_networks(
    param_size: int,
    obs_size: int,
    action_size: int,
    hidden_layer_sizes: Tuple[int, ...] = (256, 256),
) -> Tuple[networks.FeedForwardModel, networks.FeedForwardModel]:
  """Creates a policy and a value networks for SAC."""
  policy_module = networks.MLP(
      layer_sizes=hidden_layer_sizes + (param_size,),
      activation=linen.relu,
      kernel_init=jax.nn.initializers.lecun_uniform())

  class QModule(linen.Module):
    """Q Module."""
    n_critics: int = 2

    @linen.compact
    def __call__(self, obs: jnp.ndarray, actions: jnp.ndarray):
      hidden = jnp.concatenate([obs, actions], axis=-1)
      res = []
      for _ in range(self.n_critics):
        q = networks.MLP(
            layer_sizes=hidden_layer_sizes + (1,),
            activation=linen.relu,
            kernel_init=jax.nn.initializers.lecun_uniform())(
                hidden)
        res.append(q)
      return jnp.concatenate(res, axis=-1)

  q_module = QModule()

  dummy_obs = jnp.zeros((1, obs_size))
  dummy_action = jnp.zeros((1, action_size))
  policy = networks.FeedForwardModel(
      init=lambda key: policy_module.init(key, dummy_obs),
      apply=policy_module.apply)
  value = networks.FeedForwardModel(
      init=lambda key: q_module.init(key, dummy_obs, dummy_action),
      apply=q_module.apply)
  return policy, value


def train(
    environment_fn: Callable[..., envs.Env],
    num_timesteps,
    episode_length: int,
    action_repeat: int = 1,
    num_envs: int = 1,
    num_eval_envs: int = 128,
    learning_rate: float = 1e-4,
    discounting: float = 0.9,
    seed: int = 0,
    batch_size: int = 256,
    log_frequency: int = 10000,
    normalize_observations: bool = False,
    max_devices_per_host: Optional[int] = None,
    reward_scaling: float = 1.,
    tau: float = 0.005,
    min_replay_size: int = 8192,
    max_replay_size: int = 1048576,
    grad_updates_per_step: float = 1,
    progress_fn: Optional[Callable[[int, Dict[str, Any]], None]] = None,
    # The rewarder is an init function and a compute_reward function.
    # It is used to change the reward before the learner trains on it.
    make_rewarder: Optional[Callable[[], Rewarder]] = None,
    checkpoint_logdir: Optional[str] = None
):
  """SAC training."""
  assert min_replay_size % num_envs == 0
  assert max_replay_size % min_replay_size == 0
  # jax.config.update('jax_log_compiles', True)

  process_count = jax.process_count()
  process_id = jax.process_index()
  local_device_count = jax.local_device_count()
  local_devices_to_use = local_device_count
  if max_devices_per_host:
    local_devices_to_use = min(local_devices_to_use, max_devices_per_host)
  logging.info(
      'Device count: %d, process count: %d (id %d), local device count: %d, '
      'devices to be used count: %d', jax.device_count(), process_count,
      process_id, local_device_count, local_devices_to_use)

  assert max_replay_size % local_devices_to_use == 0
  assert min_replay_size % local_devices_to_use == 0
  assert num_envs % local_devices_to_use == 0

  max_replay_size = max_replay_size // local_devices_to_use
  min_replay_size = min_replay_size // local_devices_to_use

  key = jax.random.PRNGKey(seed)
  key, key_models, key_env, key_eval, key_rewarder = jax.random.split(key, 5)

  core_env = environment_fn(
      action_repeat=action_repeat,
      batch_size=num_envs // local_devices_to_use // process_count,
      episode_length=episode_length)
  key_envs = jax.random.split(key_env, local_devices_to_use)
  step_fn = core_env.step
  first_state = jax.tree_multimap(lambda *args: jnp.stack(args),
                                  *[core_env.reset(key) for key in key_envs])

  core_eval_env = environment_fn(
      action_repeat=action_repeat,
      batch_size=num_eval_envs,
      episode_length=episode_length)
  eval_first_state, eval_step_fn = env.wrap_for_eval(core_eval_env, key_eval)

  parametric_action_distribution = distribution.NormalTanhDistribution(
      event_size=core_env.action_size)

  _, obs_size = eval_first_state.core.obs.shape

  policy_model, value_model = make_sac_networks(
      parametric_action_distribution.param_size, obs_size, core_env.action_size)

  log_alpha = jnp.asarray(0., dtype=jnp.float32)
  alpha_optimizer = optax.adam(learning_rate=3e-4)
  alpha_optimizer_state = alpha_optimizer.init(log_alpha)

  policy_optimizer = optax.adam(learning_rate=learning_rate)
  q_optimizer = optax.adam(learning_rate=learning_rate)
  key_policy, key_q = jax.random.split(key_models)
  policy_params = policy_model.init(key_policy)
  policy_optimizer_state = policy_optimizer.init(policy_params)
  q_params = value_model.init(key_q)
  q_optimizer_state = q_optimizer.init(q_params)

  policy_optimizer_state, policy_params = normalization.bcast_local_devices(
      (policy_optimizer_state, policy_params), local_devices_to_use)
  q_optimizer_state, q_params = normalization.bcast_local_devices(
      (q_optimizer_state, q_params), local_devices_to_use)
  alpha_optimizer_state, log_alpha = normalization.bcast_local_devices(
      (alpha_optimizer_state, log_alpha), local_devices_to_use)

  normalizer_params, obs_normalizer_update_fn, obs_normalizer_apply_fn = (
      normalization.create_observation_normalizer(
          obs_size,
          normalize_observations,
          pmap_to_devices=local_devices_to_use))

  if make_rewarder is not None:
    init, compute_reward = make_rewarder()
    rewarder_state = init(obs_size, key_rewarder)
    rewarder_state = normalization.bcast_local_devices(rewarder_state,
                                                       local_devices_to_use)
  else:
    rewarder_state = None
    compute_reward = None

  key_debug = jax.random.PRNGKey(seed + 666)

  # EVAL
  def do_one_step_eval(carry, unused_target_t):
    state, policy_params, normalizer_params, key = carry
    key, key_sample = jax.random.split(key)
    obs = obs_normalizer_apply_fn(normalizer_params, state.core.obs)
    logits = policy_model.apply(policy_params, obs)
    actions = parametric_action_distribution.sample(logits, key_sample)
    nstate = eval_step_fn(state, actions)
    return (nstate, policy_params, normalizer_params, key), ()

  @jax.jit
  def run_eval(state, key, policy_params,
               normalizer_params) -> Tuple[env.EvalEnvState, PRNGKey]:
    policy_params, normalizer_params = jax.tree_map(
        lambda x: x[0], (policy_params, normalizer_params))
    (state, _, _, key), _ = jax.lax.scan(
        do_one_step_eval, (state, policy_params, normalizer_params, key), (),
        length=episode_length // action_repeat)
    return state, key

  # SAC
  target_entropy = -0.5 * core_env.action_size

  def alpha_loss(alpha: jnp.ndarray, policy_params: Params,
                 transitions: Transition, key: PRNGKey) -> jnp.ndarray:
    """Eq 18 from https://arxiv.org/pdf/1812.05905.pdf."""
    dist_params = policy_model.apply(policy_params, transitions.o_tm1)
    action = parametric_action_distribution.sample_no_postprocessing(
        dist_params, key)
    log_prob = parametric_action_distribution.log_prob(dist_params, action)
    alpha_loss = alpha * jax.lax.stop_gradient(-log_prob - target_entropy)
    return jnp.mean(alpha_loss)

  def critic_loss(q_params: Params, policy_params: Params,
                  target_q_params: Params, alpha: jnp.ndarray,
                  transitions: Transition, key: PRNGKey) -> jnp.ndarray:
    q_old_action = value_model.apply(q_params, transitions.o_tm1,
                                     transitions.a_tm1)
    next_dist_params = policy_model.apply(policy_params, transitions.o_t)
    next_action = parametric_action_distribution.sample_no_postprocessing(
        next_dist_params, key)
    next_log_prob = parametric_action_distribution.log_prob(
        next_dist_params, next_action)
    next_action = parametric_action_distribution.postprocess(next_action)
    next_q = value_model.apply(target_q_params, transitions.o_t, next_action)
    next_v = jnp.min(next_q, axis=-1) - alpha * next_log_prob
    target_q = jax.lax.stop_gradient(transitions.r_t * reward_scaling +
                                     transitions.d_t * discounting * next_v)
    q_error = q_old_action - jnp.expand_dims(target_q, -1)

    # Better bootstrapping for truncated episodes.
    q_error *= jnp.expand_dims(transitions.d_t, -1)

    q_loss = 0.5 * jnp.mean(jnp.square(q_error))
    return q_loss

  def actor_loss(policy_params: Params, q_params: Params,
                 alpha: jnp.ndarray, transitions: Transition,
                 key: PRNGKey) -> jnp.ndarray:
    dist_params = policy_model.apply(policy_params, transitions.o_tm1)
    action = parametric_action_distribution.sample_no_postprocessing(
        dist_params, key)
    log_prob = parametric_action_distribution.log_prob(dist_params, action)
    action = parametric_action_distribution.postprocess(action)
    q_action = value_model.apply(q_params, transitions.o_tm1, action)
    min_q = jnp.min(q_action, axis=-1)
    actor_loss = alpha * log_prob - min_q
    return jnp.mean(actor_loss)

  alpha_grad = jax.jit(jax.value_and_grad(alpha_loss))
  critic_grad = jax.jit(jax.value_and_grad(critic_loss))
  actor_grad = jax.jit(jax.value_and_grad(actor_loss))

  @jax.jit
  def update_step(
      state: TrainingState,
      transitions: jnp.ndarray,
  ) -> Tuple[TrainingState, Dict[str, jnp.ndarray]]:
    normalized_transitions = Transition(
        o_tm1=obs_normalizer_apply_fn(state.normalizer_params,
                                      transitions[:, :obs_size]),
        o_t=obs_normalizer_apply_fn(state.normalizer_params,
                                    transitions[:, obs_size:2 * obs_size]),
        a_tm1=transitions[:, 2 * obs_size:2 * obs_size + core_env.action_size],
        r_t=transitions[:, -2],
        d_t=transitions[:, -1])

    (key, key_alpha, key_critic, key_actor,
     key_rewarder) = jax.random.split(state.key, 5)

    if compute_reward is not None:
      new_rewarder_state, rewards, rewarder_metrics = compute_reward(
          state.rewarder_state, normalized_transitions, key_rewarder)
      # Assertion prevents building errors.
      assert hasattr(normalized_transitions, 'replace')
      normalized_transitions = normalized_transitions.replace(r_t=rewards)
    else:
      new_rewarder_state = state.rewarder_state
      rewarder_metrics = {}

    alpha = jnp.exp(state.alpha_params)
    alpha_loss, alpha_grads = alpha_grad(alpha, state.policy_params,
                                         normalized_transitions, key_alpha)
    critic_loss, critic_grads = critic_grad(state.q_params,
                                            state.policy_params,
                                            state.target_q_params, alpha,
                                            normalized_transitions, key_critic)
    actor_loss, actor_grads = actor_grad(state.policy_params,
                                         state.q_params, alpha,
                                         normalized_transitions, key_actor)
    alpha_grads = jax.lax.pmean(alpha_grads, axis_name='i')
    critic_grads = jax.lax.pmean(critic_grads, axis_name='i')
    actor_grads = jax.lax.pmean(actor_grads, axis_name='i')

    policy_params_update, policy_optimizer_state = policy_optimizer.update(
        actor_grads, state.policy_optimizer_state)
    policy_params = optax.apply_updates(state.policy_params,
                                        policy_params_update)
    q_params_update, q_optimizer_state = q_optimizer.update(
        critic_grads, state.q_optimizer_state)
    q_params = optax.apply_updates(state.q_params, q_params_update)
    alpha_params_update, alpha_optimizer_state = alpha_optimizer.update(
        alpha_grads, state.alpha_optimizer_state)
    alpha_params = optax.apply_updates(state.alpha_params, alpha_params_update)
    new_target_q_params = jax.tree_multimap(
        lambda x, y: x * (1 - tau) + y * tau, state.target_q_params, q_params)

    metrics = {
        'critic_loss': critic_loss,
        'actor_loss': actor_loss,
        'alpha_loss': alpha_loss,
        'alpha': jnp.exp(alpha_params),
        **rewarder_metrics
    }

    new_state = TrainingState(
        policy_optimizer_state=policy_optimizer_state,
        policy_params=policy_params,
        q_optimizer_state=q_optimizer_state,
        q_params=q_params,
        target_q_params=new_target_q_params,
        key=key,
        steps=state.steps + 1,
        alpha_optimizer_state=alpha_optimizer_state,
        alpha_params=alpha_params,
        normalizer_params=state.normalizer_params,
        rewarder_state=new_rewarder_state)
    return new_state, metrics

  def collect_data(training_state: TrainingState, state):
    key, key_sample = jax.random.split(training_state.key)
    normalized_obs = obs_normalizer_apply_fn(training_state.normalizer_params,
                                             state.obs)
    logits = policy_model.apply(training_state.policy_params, normalized_obs)
    actions = parametric_action_distribution.sample_no_postprocessing(
        logits, key_sample)
    postprocessed_actions = parametric_action_distribution.postprocess(actions)
    nstate = step_fn(state, postprocessed_actions)

    normalizer_params = obs_normalizer_update_fn(
        training_state.normalizer_params, state.obs)

    training_state = training_state.replace(
        key=key, normalizer_params=normalizer_params)

    # Concatenating data into a single data blob performs faster than 5
    # separate tensors.
    concatenated_data = jnp.concatenate([
        state.obs,
        nstate.obs,
        postprocessed_actions,
        jnp.expand_dims(nstate.reward, axis=-1),
        jnp.expand_dims(1 - nstate.done, axis=-1),
    ],
                                        axis=-1)

    return training_state, nstate, concatenated_data

  def collect_and_update_buffer(training_state, state, replay_buffer):
    training_state, state, newdata = collect_data(training_state, state)
    new_replay_data = jax.tree_multimap(
        lambda x, y: jax.lax.dynamic_update_slice_in_dim(
            x,
            y,
            replay_buffer.current_position,
            axis=0),
        replay_buffer.data,
        newdata)
    new_position = (replay_buffer.current_position +
                    num_envs // local_devices_to_use) % max_replay_size
    new_size = jnp.minimum(
        replay_buffer.current_size + num_envs // local_devices_to_use,
        max_replay_size)
    return training_state, state, ReplayBuffer(
        data=new_replay_data,
        current_position=new_position,
        current_size=new_size)

  def init_replay_buffer(training_state, state, replay_buffer):
    (training_state, state, replay_buffer), _ = jax.lax.scan(
        (lambda a, b: (collect_and_update_buffer(*a),
                       ())), (training_state, state, replay_buffer), (),
        length=min_replay_size // (num_envs // local_devices_to_use))
    return training_state, state, replay_buffer

  init_replay_buffer = jax.pmap(init_replay_buffer, axis_name='i')

  num_updates = int(num_envs * grad_updates_per_step)

  def sample_data(training_state, replay_buffer):
    key1, key2 = jax.random.split(training_state.key)
    idx = jax.random.randint(
        key2, (batch_size * num_updates // local_devices_to_use,),
        minval=0,
        maxval=replay_buffer.current_size)
    transitions = jnp.take(replay_buffer.data, idx, axis=0, mode='clip')
    transitions = jnp.reshape(transitions,
                              [num_updates, -1] + list(transitions.shape[1:]))
    training_state = training_state.replace(key=key1)
    return training_state, transitions

  def run_one_sac_epoch(carry, unused_t):
    training_state, state, replay_buffer = carry

    training_state, state, replay_buffer = collect_and_update_buffer(
        training_state, state, replay_buffer)

    training_state, transitions = sample_data(training_state, replay_buffer)
    training_state, metrics = jax.lax.scan(
        update_step, training_state, transitions, length=num_updates)

    metrics['buffer_current_size'] = replay_buffer.current_size
    metrics['buffer_current_position'] = replay_buffer.current_position
    return (training_state, state, replay_buffer), metrics

  def run_sac_training(training_state, state, replay_buffer):
    (training_state, state, replay_buffer), metrics = jax.lax.scan(
        run_one_sac_epoch, (training_state, state, replay_buffer), (),
        length=(log_frequency // action_repeat + num_envs - 1) // num_envs)
    metrics = jax.tree_map(jnp.mean, metrics)
    return training_state, state, replay_buffer, metrics

  run_sac_training = jax.pmap(run_sac_training, axis_name='i')

  training_state = TrainingState(
      policy_optimizer_state=policy_optimizer_state,
      policy_params=policy_params,
      q_optimizer_state=q_optimizer_state,
      q_params=q_params,
      target_q_params=q_params,
      key=jnp.stack(jax.random.split(key, local_devices_to_use)),
      steps=jnp.zeros((local_devices_to_use,)),
      alpha_optimizer_state=alpha_optimizer_state,
      alpha_params=log_alpha,
      normalizer_params=normalizer_params,
      rewarder_state=rewarder_state)

  training_walltime = 0
  eval_walltime = 0
  sps = 0
  eval_sps = 0
  training_metrics = {}
  state = first_state
  metrics = {}

  while True:
    current_step = int(training_state.normalizer_params[0][0]) * action_repeat
    logging.info('step %s', current_step)

    t = time.time()
    eval_state, key_debug = (
        run_eval(eval_first_state, key_debug,
                 training_state.policy_params,
                 training_state.normalizer_params))
    eval_state.completed_episodes.block_until_ready()
    eval_walltime += time.time() - t
    eval_sps = (
        episode_length * eval_first_state.core.reward.shape[0] /
        (time.time() - t))
    avg_episode_length = (
        eval_state.completed_episodes_steps / eval_state.completed_episodes)
    metrics = dict(
        dict({
            f'eval/episode_{name}': value / eval_state.completed_episodes
            for name, value in eval_state.completed_episodes_metrics.items()
        }),
        **dict({
            f'training/{name}': onp.mean(value)
            for name, value in training_metrics.items()
        }),
        **dict({
            'eval/completed_episodes': eval_state.completed_episodes,
            'eval/avg_episode_length': avg_episode_length,
            'speed/sps': sps,
            'speed/eval_sps': eval_sps,
            'speed/training_walltime': training_walltime,
            'speed/eval_walltime': eval_walltime,
            'training/grad_updates': training_state.steps[0],
        }),
    )
    logging.info(metrics)
    if progress_fn:
      progress_fn(current_step, metrics)

    if checkpoint_logdir:
      # Save current policy.
      normalizer_params = jax.tree_map(lambda x: x[0],
                                       training_state.normalizer_params)
      policy_params = jax.tree_map(lambda x: x[0],
                                   training_state.policy_params)
      params = normalizer_params, policy_params
      path = f'{checkpoint_logdir}_sac_{current_step}.flax'
      model.save_params(path, params)

    if current_step >= num_timesteps:
      break

    # Create an initialize the replay buffer.
    if current_step == 0:
      t = time.time()

      replay_buffer = ReplayBuffer(
          data=jnp.zeros((local_devices_to_use, max_replay_size,
                          obs_size * 2 + core_env.action_size + 1 + 1)),
          current_size=jnp.zeros((local_devices_to_use,), dtype=jnp.int32),
          current_position=jnp.zeros((local_devices_to_use,), dtype=jnp.int32))

      training_state, state, replay_buffer = init_replay_buffer(
          training_state, state, replay_buffer)
      training_walltime += time.time() - t

    t = time.time()
    # optimization
    training_state, state, replay_buffer, training_metrics = run_sac_training(
        training_state, state, replay_buffer)
    jax.tree_map(lambda x: x.block_until_ready(), training_metrics)
    sps = ((training_state.normalizer_params[0][0] * action_repeat -
            current_step) / (time.time() - t))
    training_walltime += time.time() - t

  normalizer_params = jax.tree_map(lambda x: x[0],
                                   training_state.normalizer_params)
  policy_params = jax.tree_map(lambda x: x[0],
                               training_state.policy_params)

  logging.info('total steps: %s', normalizer_params[0] * action_repeat)

  _, inference = make_params_and_inference_fn(core_env.observation_size,
                                              core_env.action_size,
                                              normalize_observations)
  params = normalizer_params, policy_params

  return (inference, params, metrics)


def make_params_and_inference_fn(observation_size, action_size,
                                 normalize_observations):
  """Creates params and inference function for the SAC agent."""
  obs_normalizer_params, obs_normalizer_apply_fn = normalization.make_data_and_apply_fn(
      observation_size, normalize_observations)
  parametric_action_distribution = distribution.NormalTanhDistribution(
      event_size=action_size)
  policy_model, _ = make_sac_networks(parametric_action_distribution.param_size,
                                      observation_size, action_size)

  def inference_fn(params, obs, key):
    normalizer_params, policy_params = params
    obs = obs_normalizer_apply_fn(normalizer_params, obs)
    action = parametric_action_distribution.sample(
        policy_model.apply(policy_params, obs), key)
    return action

  params = (obs_normalizer_params, policy_model.init(jax.random.PRNGKey(0)))
  return params, inference_fn



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: brax/physics/config.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='brax/physics/config.proto',
  package='brax',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x19\x62rax/physics/config.proto\x12\x04\x62rax\"*\n\x07Vector3\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\t\n\x01z\x18\x03 \x01(\x02\"W\n\x06\x46rozen\x12\x1f\n\x08position\x18\x01 \x01(\x0b\x32\r.brax.Vector3\x12\x1f\n\x08rotation\x18\x02 \x01(\x0b\x32\r.brax.Vector3\x12\x0b\n\x03\x61ll\x18\x03 \x01(\x08\"\x83\x01\n\x04\x42ody\x12\x0c\n\x04name\x18\x01 \x01(\t\x12!\n\tcolliders\x18\x02 \x03(\x0b\x32\x0e.brax.Collider\x12\x1e\n\x07inertia\x18\x03 \x01(\x0b\x32\r.brax.Vector3\x12\x0c\n\x04mass\x18\x04 \x01(\x02\x12\x1c\n\x06\x66rozen\x18\x05 \x01(\x0b\x32\x0c.brax.Frozen\"\xcd\x03\n\x08\x43ollider\x12\x1f\n\x08position\x18\x01 \x01(\x0b\x32\r.brax.Vector3\x12\x1f\n\x08rotation\x18\x02 \x01(\x0b\x32\r.brax.Vector3\x12!\n\x03\x62ox\x18\x03 \x01(\x0b\x32\x12.brax.Collider.BoxH\x00\x12%\n\x05plane\x18\x04 \x01(\x0b\x32\x14.brax.Collider.PlaneH\x00\x12\'\n\x06sphere\x18\x05 \x01(\x0b\x32\x15.brax.Collider.SphereH\x00\x12)\n\x07\x63\x61psule\x18\x06 \x01(\x0b\x32\x16.brax.Collider.CapsuleH\x00\x12-\n\theightMap\x18\x07 \x01(\x0b\x32\x18.brax.Collider.HeightMapH\x00\x1a&\n\x03\x42ox\x12\x1f\n\x08halfsize\x18\x01 \x01(\x0b\x32\r.brax.Vector3\x1a\x07\n\x05Plane\x1a\x18\n\x06Sphere\x12\x0e\n\x06radius\x18\x01 \x01(\x02\x1a\x36\n\x07\x43\x61psule\x12\x0e\n\x06radius\x18\x01 \x01(\x02\x12\x0e\n\x06length\x18\x02 \x01(\x02\x12\x0b\n\x03\x65nd\x18\x03 \x01(\x05\x1a\'\n\tHeightMap\x12\x0c\n\x04size\x18\x01 \x01(\x02\x12\x0c\n\x04\x64\x61ta\x18\x02 \x03(\x02\x42\x06\n\x04type\"\xf7\x02\n\x05Joint\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tstiffness\x18\x02 \x01(\x02\x12\x0e\n\x06parent\x18\x03 \x01(\t\x12\r\n\x05\x63hild\x18\x04 \x01(\t\x12$\n\rparent_offset\x18\x05 \x01(\x0b\x32\r.brax.Vector3\x12#\n\x0c\x63hild_offset\x18\x06 \x01(\x0b\x32\r.brax.Vector3\x12\x1f\n\x08rotation\x18\x07 \x01(\x0b\x32\r.brax.Vector3\x12\x17\n\x0f\x61ngular_damping\x18\x08 \x01(\x02\x12&\n\x0b\x61ngle_limit\x18\t \x03(\x0b\x32\x11.brax.Joint.Range\x12\x1b\n\x0elimit_strength\x18\n \x01(\x02H\x00\x88\x01\x01\x12\x1b\n\x0espring_damping\x18\x0b \x01(\x02H\x01\x88\x01\x01\x1a!\n\x05Range\x12\x0b\n\x03min\x18\x01 \x01(\x02\x12\x0b\n\x03max\x18\x02 \x01(\x02\x42\x11\n\x0f_limit_strengthB\x11\n\x0f_spring_damping\"\xa4\x01\n\x08\x41\x63tuator\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05joint\x18\x02 \x01(\t\x12\x10\n\x08strength\x18\x03 \x01(\x02\x12\'\n\x06torque\x18\x04 \x01(\x0b\x32\x15.brax.Actuator.TorqueH\x00\x12%\n\x05\x61ngle\x18\x05 \x01(\x0b\x32\x14.brax.Actuator.AngleH\x00\x1a\x08\n\x06Torque\x1a\x07\n\x05\x41ngleB\x06\n\x04type\"\xa0\x02\n\x0c\x44\x65\x66\x61ultState\x12-\n\x06\x61ngles\x18\x01 \x03(\x0b\x32\x1d.brax.DefaultState.JointAngle\x12\"\n\x03qps\x18\x02 \x03(\x0b\x32\x15.brax.DefaultState.QP\x1a\x38\n\nJointAngle\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1c\n\x05\x61ngle\x18\x02 \x01(\x0b\x32\r.brax.Vector3\x1a\x82\x01\n\x02QP\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1a\n\x03pos\x18\x02 \x01(\x0b\x32\r.brax.Vector3\x12\x1a\n\x03rot\x18\x03 \x01(\x0b\x32\r.brax.Vector3\x12\x1a\n\x03vel\x18\x04 \x01(\x0b\x32\r.brax.Vector3\x12\x1a\n\x03\x61ng\x18\x05 \x01(\x0b\x32\r.brax.Vector3\"\xb1\x03\n\x06\x43onfig\x12\x1a\n\x06\x62odies\x18\x01 \x03(\x0b\x32\n.brax.Body\x12\x1b\n\x06joints\x18\x02 \x03(\x0b\x32\x0b.brax.Joint\x12!\n\tactuators\x18\x03 \x03(\x0b\x32\x0e.brax.Actuator\x12\x12\n\nelasticity\x18\x04 \x01(\x02\x12\x10\n\x08\x66riction\x18\x05 \x01(\x02\x12\x1e\n\x07gravity\x18\x06 \x01(\x0b\x32\r.brax.Vector3\x12\x18\n\x10velocity_damping\x18\x07 \x01(\x02\x12\x17\n\x0f\x61ngular_damping\x18\x08 \x01(\x02\x12\x15\n\rbaumgarte_erp\x18\t \x01(\x02\x12.\n\x0f\x63ollide_include\x18\n \x03(\x0b\x32\x15.brax.Config.NamePair\x12\n\n\x02\x64t\x18\x0b \x01(\x02\x12\x10\n\x08substeps\x18\x0c \x01(\x05\x12\x1c\n\x06\x66rozen\x18\r \x01(\x0b\x32\x0c.brax.Frozen\x12$\n\x08\x64\x65\x66\x61ults\x18\x0e \x03(\x0b\x32\x12.brax.DefaultState\x1a)\n\x08NamePair\x12\r\n\x05\x66irst\x18\x01 \x01(\t\x12\x0e\n\x06second\x18\x02 \x01(\tb\x06proto3'
)




_VECTOR3 = _descriptor.Descriptor(
  name='Vector3',
  full_name='brax.Vector3',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='x', full_name='brax.Vector3.x', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='y', full_name='brax.Vector3.y', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='z', full_name='brax.Vector3.z', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=35,
  serialized_end=77,
)


_FROZEN = _descriptor.Descriptor(
  name='Frozen',
  full_name='brax.Frozen',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='position', full_name='brax.Frozen.position', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='rotation', full_name='brax.Frozen.rotation', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all', full_name='brax.Frozen.all', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=79,
  serialized_end=166,
)


_BODY = _descriptor.Descriptor(
  name='Body',
  full_name='brax.Body',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='brax.Body.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='colliders', full_name='brax.Body.colliders', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='inertia', full_name='brax.Body.inertia', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mass', full_name='brax.Body.mass', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='frozen', full_name='brax.Body.frozen', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=169,
  serialized_end=300,
)


_COLLIDER_BOX = _descriptor.Descriptor(
  name='Box',
  full_name='brax.Collider.Box',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='halfsize', full_name='brax.Collider.Box.halfsize', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=586,
  serialized_end=624,
)

_COLLIDER_PLANE = _descriptor.Descriptor(
  name='Plane',
  full_name='brax.Collider.Plane',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=626,
  serialized_end=633,
)

_COLLIDER_SPHERE = _descriptor.Descriptor(
  name='Sphere',
  full_name='brax.Collider.Sphere',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='radius', full_name='brax.Collider.Sphere.radius', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=635,
  serialized_end=659,
)

_COLLIDER_CAPSULE = _descriptor.Descriptor(
  name='Capsule',
  full_name='brax.Collider.Capsule',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='radius', full_name='brax.Collider.Capsule.radius', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='length', full_name='brax.Collider.Capsule.length', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end', full_name='brax.Collider.Capsule.end', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=661,
  serialized_end=715,
)

_COLLIDER_HEIGHTMAP = _descriptor.Descriptor(
  name='HeightMap',
  full_name='brax.Collider.HeightMap',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='size', full_name='brax.Collider.HeightMap.size', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='brax.Collider.HeightMap.data', index=1,
      number=2, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=717,
  serialized_end=756,
)

_COLLIDER = _descriptor.Descriptor(
  name='Collider',
  full_name='brax.Collider',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='position', full_name='brax.Collider.position', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='rotation', full_name='brax.Collider.rotation', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='box', full_name='brax.Collider.box', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='plane', full_name='brax.Collider.plane', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sphere', full_name='brax.Collider.sphere', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='capsule', full_name='brax.Collider.capsule', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='heightMap', full_name='brax.Collider.heightMap', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_COLLIDER_BOX, _COLLIDER_PLANE, _COLLIDER_SPHERE, _COLLIDER_CAPSULE, _COLLIDER_HEIGHTMAP, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='type', full_name='brax.Collider.type',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=303,
  serialized_end=764,
)


_JOINT_RANGE = _descriptor.Descriptor(
  name='Range',
  full_name='brax.Joint.Range',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='min', full_name='brax.Joint.Range.min', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max', full_name='brax.Joint.Range.max', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1071,
  serialized_end=1104,
)

_JOINT = _descriptor.Descriptor(
  name='Joint',
  full_name='brax.Joint',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='brax.Joint.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='stiffness', full_name='brax.Joint.stiffness', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parent', full_name='brax.Joint.parent', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='child', full_name='brax.Joint.child', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parent_offset', full_name='brax.Joint.parent_offset', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='child_offset', full_name='brax.Joint.child_offset', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='rotation', full_name='brax.Joint.rotation', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='angular_damping', full_name='brax.Joint.angular_damping', index=7,
      number=8, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='angle_limit', full_name='brax.Joint.angle_limit', index=8,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='limit_strength', full_name='brax.Joint.limit_strength', index=9,
      number=10, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='spring_damping', full_name='brax.Joint.spring_damping', index=10,
      number=11, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_JOINT_RANGE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_limit_strength', full_name='brax.Joint._limit_strength',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_spring_damping', full_name='brax.Joint._spring_damping',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=767,
  serialized_end=1142,
)


_ACTUATOR_TORQUE = _descriptor.Descriptor(
  name='Torque',
  full_name='brax.Actuator.Torque',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1284,
  serialized_end=1292,
)

_ACTUATOR_ANGLE = _descriptor.Descriptor(
  name='Angle',
  full_name='brax.Actuator.Angle',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1294,
  serialized_end=1301,
)

_ACTUATOR = _descriptor.Descriptor(
  name='Actuator',
  full_name='brax.Actuator',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='brax.Actuator.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='joint', full_name='brax.Actuator.joint', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='strength', full_name='brax.Actuator.strength', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='torque', full_name='brax.Actuator.torque', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='angle', full_name='brax.Actuator.angle', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_ACTUATOR_TORQUE, _ACTUATOR_ANGLE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='type', full_name='brax.Actuator.type',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=1145,
  serialized_end=1309,
)


_DEFAULTSTATE_JOINTANGLE = _descriptor.Descriptor(
  name='JointAngle',
  full_name='brax.DefaultState.JointAngle',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='brax.DefaultState.JointAngle.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='angle', full_name='brax.DefaultState.JointAngle.angle', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1411,
  serialized_end=1467,
)

_DEFAULTSTATE_QP = _descriptor.Descriptor(
  name='QP',
  full_name='brax.DefaultState.QP',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='brax.DefaultState.QP.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pos', full_name='brax.DefaultState.QP.pos', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='rot', full_name='brax.DefaultState.QP.rot', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='vel', full_name='brax.DefaultState.QP.vel', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ang', full_name='brax.DefaultState.QP.ang', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1470,
  serialized_end=1600,
)

_DEFAULTSTATE = _descriptor.Descriptor(
  name='DefaultState',
  full_name='brax.DefaultState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='angles', full_name='brax.DefaultState.angles', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='qps', full_name='brax.DefaultState.qps', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DEFAULTSTATE_JOINTANGLE, _DEFAULTSTATE_QP, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1312,
  serialized_end=1600,
)


_CONFIG_NAMEPAIR = _descriptor.Descriptor(
  name='NamePair',
  full_name='brax.Config.NamePair',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='first', full_name='brax.Config.NamePair.first', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='second', full_name='brax.Config.NamePair.second', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1995,
  serialized_end=2036,
)

_CONFIG = _descriptor.Descriptor(
  name='Config',
  full_name='brax.Config',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='bodies', full_name='brax.Config.bodies', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='joints', full_name='brax.Config.joints', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='actuators', full_name='brax.Config.actuators', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='elasticity', full_name='brax.Config.elasticity', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='friction', full_name='brax.Config.friction', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='gravity', full_name='brax.Config.gravity', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='velocity_damping', full_name='brax.Config.velocity_damping', index=6,
      number=7, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='angular_damping', full_name='brax.Config.angular_damping', index=7,
      number=8, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='baumgarte_erp', full_name='brax.Config.baumgarte_erp', index=8,
      number=9, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='collide_include', full_name='brax.Config.collide_include', index=9,
      number=10, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dt', full_name='brax.Config.dt', index=10,
      number=11, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='substeps', full_name='brax.Config.substeps', index=11,
      number=12, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='frozen', full_name='brax.Config.frozen', index=12,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='defaults', full_name='brax.Config.defaults', index=13,
      number=14, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_CONFIG_NAMEPAIR, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1603,
  serialized_end=2036,
)

_FROZEN.fields_by_name['position'].message_type = _VECTOR3
_FROZEN.fields_by_name['rotation'].message_type = _VECTOR3
_BODY.fields_by_name['colliders'].message_type = _COLLIDER
_BODY.fields_by_name['inertia'].message_type = _VECTOR3
_BODY.fields_by_name['frozen'].message_type = _FROZEN
_COLLIDER_BOX.fields_by_name['halfsize'].message_type = _VECTOR3
_COLLIDER_BOX.containing_type = _COLLIDER
_COLLIDER_PLANE.containing_type = _COLLIDER
_COLLIDER_SPHERE.containing_type = _COLLIDER
_COLLIDER_CAPSULE.containing_type = _COLLIDER
_COLLIDER_HEIGHTMAP.containing_type = _COLLIDER
_COLLIDER.fields_by_name['position'].message_type = _VECTOR3
_COLLIDER.fields_by_name['rotation'].message_type = _VECTOR3
_COLLIDER.fields_by_name['box'].message_type = _COLLIDER_BOX
_COLLIDER.fields_by_name['plane'].message_type = _COLLIDER_PLANE
_COLLIDER.fields_by_name['sphere'].message_type = _COLLIDER_SPHERE
_COLLIDER.fields_by_name['capsule'].message_type = _COLLIDER_CAPSULE
_COLLIDER.fields_by_name['heightMap'].message_type = _COLLIDER_HEIGHTMAP
_COLLIDER.oneofs_by_name['type'].fields.append(
  _COLLIDER.fields_by_name['box'])
_COLLIDER.fields_by_name['box'].containing_oneof = _COLLIDER.oneofs_by_name['type']
_COLLIDER.oneofs_by_name['type'].fields.append(
  _COLLIDER.fields_by_name['plane'])
_COLLIDER.fields_by_name['plane'].containing_oneof = _COLLIDER.oneofs_by_name['type']
_COLLIDER.oneofs_by_name['type'].fields.append(
  _COLLIDER.fields_by_name['sphere'])
_COLLIDER.fields_by_name['sphere'].containing_oneof = _COLLIDER.oneofs_by_name['type']
_COLLIDER.oneofs_by_name['type'].fields.append(
  _COLLIDER.fields_by_name['capsule'])
_COLLIDER.fields_by_name['capsule'].containing_oneof = _COLLIDER.oneofs_by_name['type']
_COLLIDER.oneofs_by_name['type'].fields.append(
  _COLLIDER.fields_by_name['heightMap'])
_COLLIDER.fields_by_name['heightMap'].containing_oneof = _COLLIDER.oneofs_by_name['type']
_JOINT_RANGE.containing_type = _JOINT
_JOINT.fields_by_name['parent_offset'].message_type = _VECTOR3
_JOINT.fields_by_name['child_offset'].message_type = _VECTOR3
_JOINT.fields_by_name['rotation'].message_type = _VECTOR3
_JOINT.fields_by_name['angle_limit'].message_type = _JOINT_RANGE
_JOINT.oneofs_by_name['_limit_strength'].fields.append(
  _JOINT.fields_by_name['limit_strength'])
_JOINT.fields_by_name['limit_strength'].containing_oneof = _JOINT.oneofs_by_name['_limit_strength']
_JOINT.oneofs_by_name['_spring_damping'].fields.append(
  _JOINT.fields_by_name['spring_damping'])
_JOINT.fields_by_name['spring_damping'].containing_oneof = _JOINT.oneofs_by_name['_spring_damping']
_ACTUATOR_TORQUE.containing_type = _ACTUATOR
_ACTUATOR_ANGLE.containing_type = _ACTUATOR
_ACTUATOR.fields_by_name['torque'].message_type = _ACTUATOR_TORQUE
_ACTUATOR.fields_by_name['angle'].message_type = _ACTUATOR_ANGLE
_ACTUATOR.oneofs_by_name['type'].fields.append(
  _ACTUATOR.fields_by_name['torque'])
_ACTUATOR.fields_by_name['torque'].containing_oneof = _ACTUATOR.oneofs_by_name['type']
_ACTUATOR.oneofs_by_name['type'].fields.append(
  _ACTUATOR.fields_by_name['angle'])
_ACTUATOR.fields_by_name['angle'].containing_oneof = _ACTUATOR.oneofs_by_name['type']
_DEFAULTSTATE_JOINTANGLE.fields_by_name['angle'].message_type = _VECTOR3
_DEFAULTSTATE_JOINTANGLE.containing_type = _DEFAULTSTATE
_DEFAULTSTATE_QP.fields_by_name['pos'].message_type = _VECTOR3
_DEFAULTSTATE_QP.fields_by_name['rot'].message_type = _VECTOR3
_DEFAULTSTATE_QP.fields_by_name['vel'].message_type = _VECTOR3
_DEFAULTSTATE_QP.fields_by_name['ang'].message_type = _VECTOR3
_DEFAULTSTATE_QP.containing_type = _DEFAULTSTATE
_DEFAULTSTATE.fields_by_name['angles'].message_type = _DEFAULTSTATE_JOINTANGLE
_DEFAULTSTATE.fields_by_name['qps'].message_type = _DEFAULTSTATE_QP
_CONFIG_NAMEPAIR.containing_type = _CONFIG
_CONFIG.fields_by_name['bodies'].message_type = _BODY
_CONFIG.fields_by_name['joints'].message_type = _JOINT
_CONFIG.fields_by_name['actuators'].message_type = _ACTUATOR
_CONFIG.fields_by_name['gravity'].message_type = _VECTOR3
_CONFIG.fields_by_name['collide_include'].message_type = _CONFIG_NAMEPAIR
_CONFIG.fields_by_name['frozen'].message_type = _FROZEN
_CONFIG.fields_by_name['defaults'].message_type = _DEFAULTSTATE
DESCRIPTOR.message_types_by_name['Vector3'] = _VECTOR3
DESCRIPTOR.message_types_by_name['Frozen'] = _FROZEN
DESCRIPTOR.message_types_by_name['Body'] = _BODY
DESCRIPTOR.message_types_by_name['Collider'] = _COLLIDER
DESCRIPTOR.message_types_by_name['Joint'] = _JOINT
DESCRIPTOR.message_types_by_name['Actuator'] = _ACTUATOR
DESCRIPTOR.message_types_by_name['DefaultState'] = _DEFAULTSTATE
DESCRIPTOR.message_types_by_name['Config'] = _CONFIG
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Vector3 = _reflection.GeneratedProtocolMessageType('Vector3', (_message.Message,), {
  'DESCRIPTOR' : _VECTOR3,
  '__module__' : 'brax.physics.config_pb2'
  # @@protoc_insertion_point(class_scope:brax.Vector3)
  })
_sym_db.RegisterMessage(Vector3)

Frozen = _reflection.GeneratedProtocolMessageType('Frozen', (_message.Message,), {
  'DESCRIPTOR' : _FROZEN,
  '__module__' : 'brax.physics.config_pb2'
  # @@protoc_insertion_point(class_scope:brax.Frozen)
  })
_sym_db.RegisterMessage(Frozen)

Body = _reflection.GeneratedProtocolMessageType('Body', (_message.Message,), {
  'DESCRIPTOR' : _BODY,
  '__module__' : 'brax.physics.config_pb2'
  # @@protoc_insertion_point(class_scope:brax.Body)
  })
_sym_db.RegisterMessage(Body)

Collider = _reflection.GeneratedProtocolMessageType('Collider', (_message.Message,), {

  'Box' : _reflection.GeneratedProtocolMessageType('Box', (_message.Message,), {
    'DESCRIPTOR' : _COLLIDER_BOX,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Collider.Box)
    })
  ,

  'Plane' : _reflection.GeneratedProtocolMessageType('Plane', (_message.Message,), {
    'DESCRIPTOR' : _COLLIDER_PLANE,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Collider.Plane)
    })
  ,

  'Sphere' : _reflection.GeneratedProtocolMessageType('Sphere', (_message.Message,), {
    'DESCRIPTOR' : _COLLIDER_SPHERE,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Collider.Sphere)
    })
  ,

  'Capsule' : _reflection.GeneratedProtocolMessageType('Capsule', (_message.Message,), {
    'DESCRIPTOR' : _COLLIDER_CAPSULE,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Collider.Capsule)
    })
  ,

  'HeightMap' : _reflection.GeneratedProtocolMessageType('HeightMap', (_message.Message,), {
    'DESCRIPTOR' : _COLLIDER_HEIGHTMAP,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Collider.HeightMap)
    })
  ,
  'DESCRIPTOR' : _COLLIDER,
  '__module__' : 'brax.physics.config_pb2'
  # @@protoc_insertion_point(class_scope:brax.Collider)
  })
_sym_db.RegisterMessage(Collider)
_sym_db.RegisterMessage(Collider.Box)
_sym_db.RegisterMessage(Collider.Plane)
_sym_db.RegisterMessage(Collider.Sphere)
_sym_db.RegisterMessage(Collider.Capsule)
_sym_db.RegisterMessage(Collider.HeightMap)

Joint = _reflection.GeneratedProtocolMessageType('Joint', (_message.Message,), {

  'Range' : _reflection.GeneratedProtocolMessageType('Range', (_message.Message,), {
    'DESCRIPTOR' : _JOINT_RANGE,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Joint.Range)
    })
  ,
  'DESCRIPTOR' : _JOINT,
  '__module__' : 'brax.physics.config_pb2'
  # @@protoc_insertion_point(class_scope:brax.Joint)
  })
_sym_db.RegisterMessage(Joint)
_sym_db.RegisterMessage(Joint.Range)

Actuator = _reflection.GeneratedProtocolMessageType('Actuator', (_message.Message,), {

  'Torque' : _reflection.GeneratedProtocolMessageType('Torque', (_message.Message,), {
    'DESCRIPTOR' : _ACTUATOR_TORQUE,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Actuator.Torque)
    })
  ,

  'Angle' : _reflection.GeneratedProtocolMessageType('Angle', (_message.Message,), {
    'DESCRIPTOR' : _ACTUATOR_ANGLE,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Actuator.Angle)
    })
  ,
  'DESCRIPTOR' : _ACTUATOR,
  '__module__' : 'brax.physics.config_pb2'
  # @@protoc_insertion_point(class_scope:brax.Actuator)
  })
_sym_db.RegisterMessage(Actuator)
_sym_db.RegisterMessage(Actuator.Torque)
_sym_db.RegisterMessage(Actuator.Angle)

DefaultState = _reflection.GeneratedProtocolMessageType('DefaultState', (_message.Message,), {

  'JointAngle' : _reflection.GeneratedProtocolMessageType('JointAngle', (_message.Message,), {
    'DESCRIPTOR' : _DEFAULTSTATE_JOINTANGLE,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.DefaultState.JointAngle)
    })
  ,

  'QP' : _reflection.GeneratedProtocolMessageType('QP', (_message.Message,), {
    'DESCRIPTOR' : _DEFAULTSTATE_QP,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.DefaultState.QP)
    })
  ,
  'DESCRIPTOR' : _DEFAULTSTATE,
  '__module__' : 'brax.physics.config_pb2'
  # @@protoc_insertion_point(class_scope:brax.DefaultState)
  })
_sym_db.RegisterMessage(DefaultState)
_sym_db.RegisterMessage(DefaultState.JointAngle)
_sym_db.RegisterMessage(DefaultState.QP)

Config = _reflection.GeneratedProtocolMessageType('Config', (_message.Message,), {

  'NamePair' : _reflection.GeneratedProtocolMessageType('NamePair', (_message.Message,), {
    'DESCRIPTOR' : _CONFIG_NAMEPAIR,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Config.NamePair)
    })
  ,
  'DESCRIPTOR' : _CONFIG,
  '__module__' : 'brax.physics.config_pb2'
  # @@protoc_insertion_point(class_scope:brax.Config)
  })
_sym_db.RegisterMessage(Config)
_sym_db.RegisterMessage(Config.NamePair)


# @@protoc_insertion_point(module_scope)



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Some general purpose math functions."""

from typing import Tuple

import jax.numpy as jnp

from brax.physics.base import QP


def safe_norm(d, axis=None):
  """Calculates a jnp.linalg.norm(d) that's safe for gradients at d=0.

  These gymnastics are to avoid a poorly defined gradient for jnp.linal.norm(0)
  see https://github.com/google/jax/issues/3058 for details

  Args:
    d: A jnp.array
    axis: The axis along which to compute the norm

  Returns:
    Equivalent to jnp.linalg.norm(d)
  """
  is_zero = jnp.allclose(d, 0.)
  d = jnp.where(is_zero, jnp.ones_like(d), d)  # replace d with ones if is_zero
  l = jnp.linalg.norm(d, axis=axis)
  l = jnp.where(is_zero, 0., l)  # replace norm with zero if is_zero

  return l


def rotate(vec: jnp.ndarray, quat: jnp.ndarray):
  """Rotates a vector vec by a unit quaternion quat.

  Args:
    vec: jnp.ndarray (3)
    quat: jnp.ndarray (4) (w,x,y,z)

  Returns:
    A jnp.ndarry(3) containing vec rotated by quat.
  """

  u = quat[1:]
  s = quat[0]
  return 2 * (jnp.dot(u, vec) *
              u) + (s * s - jnp.dot(u, u)) * vec + 2 * s * jnp.cross(u, vec)


def inv_rotate(vec: jnp.ndarray, quat: jnp.ndarray):
  """Rotates a vector by the inverse of a unit quaternion.

  Args:
    vec: jnp.ndarray
    quat: jnp.ndarray

  Returns:
    A vector rotated by quat^{-1}
  """
  u = -1. * quat[1:]
  s = quat[0]
  return 2 * (jnp.dot(u, vec) *
              u) + (s * s - jnp.dot(u, u)) * vec + 2 * s * jnp.cross(u, vec)


def ang_to_quat(ang: jnp.ndarray):
  """Converts angular velocity to a quaternion.

  Args:
    ang: angular velocity

  Returns:
    A rotation quaternion.
  """
  return jnp.array([[0., -ang[0], -ang[1], -ang[2]],
                    [ang[0], 0, -ang[2], ang[1]], [ang[1], ang[2], 0., -ang[0]],
                    [ang[2], -ang[1], ang[0], 0.]])


def quat_to_axis_angle(q: jnp.ndarray):
  """Returns the axis-angle representation of a quaternion.

  Args:
    q: (4,) a quaternion

  Returns:
    The angle of axis-angle of this quaternion, in the range [-pi, pi].
  """
  # TODO: replace with more accurate safe function
  # avoid the singularity at 0:
  epsilon = 0.00001
  # safety 1e-6 jitter added because both sqrt and arctan2 have bad gradients
  denom = jnp.sqrt(epsilon + 1 - q[0] * q[0])
  angle = 2. * jnp.arctan2(
      jnp.sqrt(epsilon + q[1] * q[1] + q[2] * q[2] + q[3] * q[3]), q[0])
  angle += jnp.where(angle > jnp.pi, x=-2. * jnp.pi, y=0.)
  angle += jnp.where(angle < -jnp.pi, x=2. * jnp.pi, y=0.)
  return jnp.array([q[1] / denom, q[2] / denom, q[3] / denom]), angle


def to_world(qp: QP, rpos: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
  """Returns world information about a point relative to a part.

  Args:
    qp: Part from which to offset point.
    rpos: Point relative to center of mass of part.

  Returns:
    A 2-tuple containing:
      * World-space coordinates of rpos
      * World-space velocity of rpos
  """
  rpos_off = rotate(rpos, qp.rot)
  rvel = jnp.cross(qp.ang, rpos_off)
  return (qp.pos + rpos_off, qp.vel + rvel)


def world_velocity(qp: QP, pos: jnp.ndarray) -> jnp.ndarray:
  """Returns the velocity of the point on a rigidbody in world space.

  Args:
    qp: Part from which to extract a velocity.
    pos: World space position which to use for velocity calculation.
  """
  return qp.vel + jnp.cross(qp.ang, pos - qp.pos)


def signed_angle(qp_p: QP, qp_c: QP, normal_axis: jnp.ndarray,
                 ref_vector) -> float:
  """Calculates the signed angle between two parts.

  This calculation works by checking how much a parent and child part's quats
  rotate a reference vector.

  Args:
    qp_p: State data for the parent part
    qp_c: State data for the child part
    normal_axis: Common axis around which to calculate change in angle
    ref_vector: A reference vector pointing at 0-degrees offset

  Returns:
    The signed angle between two parts.
  """
  ref_vector_p = rotate(ref_vector, qp_p.rot)
  ref_vector_c = rotate(ref_vector, qp_c.rot)
  angle = jnp.arctan2(
      jnp.dot(jnp.cross(ref_vector_p, ref_vector_c), normal_axis),
      jnp.dot(ref_vector_p, ref_vector_c))
  return angle


def qmult(u, v):
  """Multiplies two quaternions.

  Args:
    u: jnp.ndarray (4) (w,x,y,z)
    v: jnp.ndarray (4) (w,x,y,z)

  Returns:
    A quaternion u*v.
  """
  return jnp.array([
      u[0] * v[0] - u[1] * v[1] - u[2] * v[2] - u[3] * v[3],
      u[0] * v[1] + u[1] * v[0] + u[2] * v[3] - u[3] * v[2],
      u[0] * v[2] - u[1] * v[3] + u[2] * v[0] + u[3] * v[1],
      u[0] * v[3] + u[1] * v[2] - u[2] * v[1] + u[3] * v[0],
  ])


def quat_rot_axis(v, angle):
  """Provides a quaternion that describes rotating around axis v by angle.

  Args:
    v: jnp.ndarray (3) (x,y,z)
    angle: float angle to rotate by

  Returns:
    A quaternion that rotates around v by angle
  """
  qx = v[0] * jnp.sin(angle / 2.)
  qy = v[1] * jnp.sin(angle / 2.)
  qz = v[2] * jnp.sin(angle / 2.)
  qw = jnp.cos(angle / 2.)
  return jnp.array([qw, qx, qy, qz])


def quat_rot_between_vec(u, v):
  """Provides quaternion that describes rotation from u to v.

  Args:
    u: jnp.ndarray (3) (x,y,z)
    v: jnp.ndarray (4) (x,y,z)

  Returns:
    A quaternion describing rotation from u to v.
  """
  angle = jnp.arccos(
      jnp.vdot(u, v) / jnp.sqrt(jnp.dot(u, u)) / jnp.sqrt(jnp.dot(v, v)))
  direction = jnp.cross(u, v)
  direction /= jnp.sqrt(jnp.vdot(direction, direction))
  direction *= jnp.sin(angle / 2.)
  return jnp.array([
      jnp.cos(angle / 2.),
      direction[0],
      direction[1],
      direction[2],
  ])


def quat_rot_between_xy_vec(u, v):
  """Provides quaternion that describes rotation from u to v around the z axis.

  Args:
    u: jnp.ndarray (2) (x,y)
    v: jnp.ndarray (2) (x,y)

  Returns:
    A quaternion describing rotation from u to v.
  """
  angle = jnp.arctan2(v[1], v[0]) - jnp.arctan2(u[1], u[0])
  direction = jnp.array([0., 0., 1.])
  direction *= jnp.sin(angle / 2.)
  return jnp.array([
      jnp.cos(jnp.abs(angle) / 2.),
      direction[0],
      direction[1],
      direction[2],
  ])


# TODO: discern from jax team why this is faster than linalg.det
def det(r1, r2, r3):
  """Calculates the determinant of a 3x3 matrix with rows as args.

  Args:
    r1: First row
    r2: Second row
    r3: Third row

  Returns:
    Determinant of matrix [r1,r2,r3].  Functionally equivalent to
    jnp.linalg.det(jnp.array([r1,r2,r3])), but jits 10x faster for large batch.

  """
  return r1[0] * r2[1] * r3[2] + r1[1] * r2[2] * r3[0] + r1[2] * r2[0] * r3[
      1] - r1[2] * r2[1] * r3[0] - r1[0] * r2[2] * r3[1] - r1[1] * r2[0] * r3[2]


def relative_quat(quat_1, quat_2):
  """Calculates the relative quaternion rotation from quat_1 to quat_2.

  Args:
    quat_1: First quaterion
    quat_2: Second quaternion

  Returns:
    A quaternion that is equivalent to the rotation achieved by inverting quat_1
    and then applying quat_2.
  """
  inv_quat_1 = inv_quat(quat_1)
  new_quat = qmult(quat_2, inv_quat_1)
  return new_quat


def inv_quat(q):
  """Calculates the inverse of quaternion q.

  Args:
    q: Some quaternion [w, x, y, z]

  Returns:
    The inverse of q, where qmult(q, inv_quat(q)) = [1, 0, 0, 0].
  """
  return q * jnp.array([1., -1., -1., -1.])



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Analytic policy gradient training."""

import time
from typing import Any, Callable, Dict, Optional, Tuple

from absl import logging
from brax import envs
from brax.training import distribution
from brax.training import env
from brax.training import networks
from brax.training import normalization
from brax.training.types import Params
from brax.training.types import PRNGKey
import flax
from flax import linen
import jax
import jax.numpy as jnp
import optax


@flax.struct.dataclass
class TrainingState:
  """Contains training state for the learner."""
  key: PRNGKey
  normalizer_params: Params
  optimizer_state: optax.OptState
  policy_params: Params


def train(
    environment_fn: Callable[..., envs.Env],
    episode_length: int,
    action_repeat: int = 1,
    num_envs: int = 1,
    num_eval_envs: int = 128,
    max_gradient_norm: float = 1e9,
    max_devices_per_host: Optional[int] = None,
    learning_rate=1e-4,
    normalize_observations=False,
    seed=0,
    log_frequency=10,
    progress_fn: Optional[Callable[[int, Dict[str, Any]], None]] = None,
    truncation_length: Optional[int] = None,
):
  """Direct trajectory optimization training."""
  xt = time.time()

  process_count = jax.process_count()
  process_id = jax.process_index()
  local_device_count = jax.local_device_count()
  local_devices_to_use = local_device_count
  if max_devices_per_host:
    local_devices_to_use = min(local_devices_to_use, max_devices_per_host)
  logging.info(
      'Device count: %d, process count: %d (id %d), local device count: %d, '
      'devices to be used count: %d', jax.device_count(), process_count,
      process_id, local_device_count, local_devices_to_use)

  key = jax.random.PRNGKey(seed)
  key, key_models, key_env = jax.random.split(key, 3)

  key_env = jax.random.split(key_env, process_count)[process_id]
  key = jax.random.split(key, process_count)[process_id]

  core_env = environment_fn(
      action_repeat=action_repeat,
      batch_size=num_envs // local_devices_to_use // process_count,
      episode_length=episode_length)
  key_envs = jax.random.split(key_env, local_devices_to_use)
  step_fn = core_env.step
  first_state = jax.tree_multimap(lambda *args: jnp.stack(args),
                                  *[core_env.reset(key) for key in key_envs])

  core_eval_env = environment_fn(
      action_repeat=action_repeat,
      batch_size=num_eval_envs,
      episode_length=episode_length)
  eval_first_state, eval_step_fn = env.wrap_for_eval(core_eval_env, key_env)

  parametric_action_distribution = distribution.NormalTanhDistribution(
      event_size=core_env.action_size)

  policy_model = make_direct_optimization_model(parametric_action_distribution,
                                                core_env.observation_size)

  policy_params = policy_model.init(key_models)
  optimizer = optax.adam(learning_rate=learning_rate)
  optimizer_state = optimizer.init(policy_params)
  optimizer_state, policy_params = normalization.bcast_local_devices(
      (optimizer_state, policy_params), local_devices_to_use)

  normalizer_params, obs_normalizer_update_fn, obs_normalizer_apply_fn = (
      normalization.create_observation_normalizer(
          core_env.observation_size,
          normalize_observations,
          num_leading_batch_dims=2,
          pmap_to_devices=local_devices_to_use))

  key_debug = jax.random.PRNGKey(seed + 666)

  def do_one_step_eval(carry, unused_target_t):
    state, params, normalizer_params, key = carry
    key, key_sample = jax.random.split(key)
    # TODO: Make this nicer ([0] comes from pmapping).
    obs = obs_normalizer_apply_fn(
        jax.tree_map(lambda x: x[0], normalizer_params), state.core.obs)
    print(obs.shape)
    print(jax.tree_map(lambda x: x.shape, params))
    logits = policy_model.apply(params, obs)
    actions = parametric_action_distribution.sample(logits, key_sample)
    nstate = eval_step_fn(state, actions)
    return (nstate, params, normalizer_params, key), ()

  @jax.jit
  def run_eval(params, state, normalizer_params,
               key) -> Tuple[env.EvalEnvState, PRNGKey]:
    params = jax.tree_map(lambda x: x[0], params)
    (state, _, _, key), _ = jax.lax.scan(
        do_one_step_eval, (state, params, normalizer_params, key), (),
        length=episode_length // action_repeat)
    return state, key

  def do_one_step(carry, step_index):
    state, params, normalizer_params, key = carry
    key, key_sample = jax.random.split(key)
    normalized_obs = obs_normalizer_apply_fn(normalizer_params, state.obs)
    logits = policy_model.apply(params, normalized_obs)
    actions = parametric_action_distribution.sample(logits, key_sample)
    nstate = step_fn(state, actions)
    if truncation_length is not None and truncation_length > 0:
      nstate = jax.lax.cond(
          jnp.mod(step_index + 1, truncation_length) == 0.,
          jax.lax.stop_gradient, lambda x: x, nstate)

    return (nstate, params, normalizer_params, key), (nstate.reward,
                                                      state.obs)

  def loss(params, normalizer_params, state, key):
    _, (rewards, obs) = jax.lax.scan(
        do_one_step, (state, params, normalizer_params, key),
        (jnp.array(range(episode_length // action_repeat))),
        length=episode_length // action_repeat)
    normalizer_params = obs_normalizer_update_fn(normalizer_params, obs)
    return -jnp.mean(rewards), normalizer_params

  loss_grad = jax.grad(loss, has_aux=True)

  def clip_by_global_norm(updates):
    g_norm = optax.global_norm(updates)
    trigger = g_norm < max_gradient_norm
    updates = jax.tree_multimap(
        lambda t: jnp.where(trigger, t, (t / g_norm) * max_gradient_norm),
        updates)
    return updates

  def _minimize(training_state: TrainingState, state: envs.State):
    key, key_grad = jax.random.split(training_state.key)
    grad, normalizer_params = loss_grad(training_state.policy_params,
                                        training_state.normalizer_params,
                                        state, key_grad)
    grad = clip_by_global_norm(grad)
    grad = jax.lax.pmean(grad, axis_name='i')
    params_update, optimizer_state = optimizer.update(
        grad, training_state.optimizer_state)
    policy_params = optax.apply_updates(training_state.policy_params,
                                        params_update)

    metrics = {
        'grad_norm': optax.global_norm(grad),
        'params_norm': optax.global_norm(policy_params)
    }
    return TrainingState(
        key=key,
        optimizer_state=optimizer_state,
        normalizer_params=normalizer_params,
        policy_params=policy_params), metrics

  minimize = jax.pmap(_minimize, axis_name='i')

  logging.info('Available devices %s', jax.devices())
  training_walltime = 0
  sps = 0
  eval_sps = 0
  summary = {
      'params_norm':
          optax.global_norm(jax.tree_map(lambda x: x[0], policy_params))
  }
  key = jnp.stack(jax.random.split(key, local_devices_to_use))
  training_state = TrainingState(key=key,
                                 optimizer_state=optimizer_state,
                                 normalizer_params=normalizer_params,
                                 policy_params=policy_params)

  for it in range(log_frequency + 1):
    logging.info('starting iteration %s %s', it, time.time() - xt)
    t = time.time()

    if process_id == 0:
      eval_state, key_debug = run_eval(training_state.policy_params,
                                       eval_first_state,
                                       training_state.normalizer_params,
                                       key_debug)
      eval_state.completed_episodes.block_until_ready()
      eval_sps = (
          episode_length * eval_first_state.core.reward.shape[0] /
          (time.time() - t))
      avg_episode_length = (
          eval_state.completed_episodes_steps / eval_state.completed_episodes)
      metrics = dict(
          dict({
              f'eval/episode_{name}': value / eval_state.completed_episodes
              for name, value in eval_state.completed_episodes_metrics.items()
          }),
          **dict({
              'eval/completed_episodes': eval_state.completed_episodes,
              'eval/avg_episode_length': avg_episode_length,
              'speed/sps': sps,
              'speed/eval_sps': eval_sps,
              'speed/training_walltime': training_walltime,
              'speed/timestamp': training_walltime,
              'train/grad_norm': jnp.mean(summary.get('grad_norm', 0)),
              'train/params_norm': jnp.mean(summary.get('params_norm', 0)),
          }))

      logging.info(metrics)
      if progress_fn:
        progress_fn(it, metrics)

    if it == log_frequency:
      break

    t = time.time()
    # optimization
    training_state, summary = minimize(training_state, first_state)
    jax.tree_map(lambda x: x.block_until_ready(), summary)
    sps = (episode_length * num_envs) / (time.time() - t)
    training_walltime += time.time() - t

  params = jax.tree_map(lambda x: x[0], training_state.policy_params)
  normalizer_params = jax.tree_map(lambda x: x[0],
                                   training_state.normalizer_params)
  params = normalizer_params, params
  _, inference = make_params_and_inference_fn(core_env.observation_size,
                                              core_env.action_size,
                                              normalize_observations)

  if process_count > 1:
    # Make sure all processes stay up until the end of main.
    x = jnp.ones([jax.local_device_count()])
    x = jax.device_get(jax.pmap(lambda x: jax.lax.psum(x, 'i'), 'i')(x))
    assert x[0] == jax.device_count()

  return (inference, params, metrics)


def make_direct_optimization_model(parametric_action_distribution, obs_size):
  return networks.make_model(
      [32, 32, 32, 32, parametric_action_distribution.param_size],
      obs_size,
      activation=linen.swish)


def make_params_and_inference_fn(observation_size, action_size,
                                 normalize_observations):
  """Creates params and inference function for the direct optimization agent."""
  parametric_action_distribution = distribution.NormalTanhDistribution(
      event_size=action_size)
  obs_normalizer_params, obs_normalizer_apply_fn = normalization.make_data_and_apply_fn(
      observation_size, normalize_observations)
  policy_model = make_direct_optimization_model(parametric_action_distribution,
                                                observation_size)

  def inference_fn(params, obs, key):
    normalizer_params, params = params
    obs = obs_normalizer_apply_fn(normalizer_params, obs)
    action = parametric_action_distribution.sample(
        policy_model.apply(params, obs), key)
    return action

  params = (obs_normalizer_params, policy_model.init(jax.random.PRNGKey(0)))
  return params, inference_fn



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""A brax system."""

import functools
from typing import Tuple

import jax
import jax.numpy as jnp
from brax.physics import actuators
from brax.physics import bodies
from brax.physics import colliders
from brax.physics import config_pb2
from brax.physics import integrators
from brax.physics import joints
from brax.physics import math
from brax.physics import tree
from brax.physics.base import Info, P, Q, QP, euler_to_quat, validate_config, vec_to_np


class System:
  """A brax system."""

  def __init__(self, config: config_pb2.Config):
    self.config = validate_config(config)

    self.num_bodies = len(config.bodies)
    self.body_idx = {b.name: i for i, b in enumerate(config.bodies)}

    self.active_pos = 1. * jnp.logical_not(
        jnp.array([vec_to_np(b.frozen.position) for b in config.bodies]))
    self.active_rot = 1. * jnp.logical_not(
        jnp.array([vec_to_np(b.frozen.rotation) for b in config.bodies]))

    self.box_plane = colliders.BoxPlane(config)
    # TODO: add capsule <> heightmap
    self.box_heightMap = colliders.BoxHeightMap(config)
    self.capsule_plane = colliders.CapsulePlane(config)
    self.capsule_capsule = colliders.CapsuleCapsule(config)

    self.num_joints = len(config.joints)
    self.joint_revolute = joints.Revolute.from_config(config)
    self.joint_universal = joints.Universal.from_config(config)
    self.joint_spherical = joints.Spherical.from_config(config)

    self.num_actuators = len(config.actuators)
    self.num_joint_dof = sum(len(j.angle_limit) for j in config.joints)

    self.angle_1d = actuators.Angle.from_config(config, self.joint_revolute)
    self.angle_2d = actuators.Angle.from_config(config, self.joint_universal)
    self.angle_3d = actuators.Angle.from_config(config, self.joint_spherical)
    self.torque_1d = actuators.Torque.from_config(config, self.joint_revolute)
    self.torque_2d = actuators.Torque.from_config(config, self.joint_universal)
    self.torque_3d = actuators.Torque.from_config(config, self.joint_spherical)

  @functools.partial(jax.jit, static_argnums=(0, 1))
  def default_qp(self, default_index: int = 0) -> QP:
    """Returns a default state for the system."""
    qps = {}

    # do we have default overrides in the config?
    default_angles, default_qps = {}, {}
    if default_index < len(self.config.defaults):
      defaults = self.config.defaults[default_index]
      default_qps = {qp.name: qp for qp in defaults.qps}
      default_angles = {angle.name: angle for angle in defaults.angles}

    # make the kinematic tree of bodies
    root = tree.Node.from_config(self.config)

    parent_joint = {j.child: j for j in self.config.joints}
    rot_axes = jax.jit(jax.vmap(math.rotate, in_axes=[0, None]))

    for body in root.depth_first():
      if body.name in default_qps:
        # set qp if found in defaults.qps override
        qp = default_qps[body.name]
        qps[body.name] = QP(
            pos=vec_to_np(qp.pos),
            rot=euler_to_quat(qp.rot),
            vel=vec_to_np(qp.vel),
            ang=vec_to_np(qp.ang))
      elif body.name in parent_joint:
        # pos/rot can be determined if the body is the child of a joint
        joint = parent_joint[body.name]
        axes = rot_axes(jnp.eye(3), euler_to_quat(joint.rotation))
        if joint.name in default_angles:
          angles = vec_to_np(default_angles[joint.name].angle) * jnp.pi / 180
        else:
          angles = [(l.min + l.max) * jnp.pi / 360 for l in joint.angle_limit]

        # for each joint angle, rotate by that angle.
        # these are euler intrinsic rotations, so the axes are rotated too
        local_rot = jnp.array([1., 0., 0., 0.])
        for axis, angle in zip(axes, angles):
          axis_rotated = math.rotate(axis, local_rot)
          next_rot = math.quat_rot_axis(axis_rotated, angle)
          local_rot = math.qmult(next_rot, local_rot)

        local_offset = math.rotate(vec_to_np(joint.child_offset), local_rot)
        local_pos = vec_to_np(joint.parent_offset) - local_offset

        parent_qp = qps[joint.parent]
        rot = math.qmult(parent_qp.rot, local_rot)
        pos = parent_qp.pos + math.rotate(local_pos, parent_qp.rot)
        qps[body.name] = QP(
            pos=pos, rot=rot, vel=jnp.zeros(3), ang=jnp.zeros(3))
      else:
        qps[body.name] = QP.zero()

    # any trees that have no body qp overrides in the config are set just above
    # the xy plane.  this convenience operation may be removed in the future.
    body_map = {body.name: body for body in self.config.bodies}
    for node in root.children:
      children = [node.name] + [n.name for n in node.depth_first()]
      if any(c in default_qps for c in children):
        continue  # ignore a tree if some part of it is overriden

      zs = jnp.array([bodies.min_z(qps[c], body_map[c]) for c in children])
      min_z = jnp.min(zs)
      for body in children:
        qp = qps[body]
        pos = qp.pos - min_z * jnp.array([0., 0., 1.])
        qps[body] = qp.replace(pos=pos)

    qps = [qps[body.name] for body in self.config.bodies]
    return jax.tree_multimap((lambda *args: jnp.stack(args)), *qps)

  @functools.partial(jax.jit, static_argnums=(0,))
  def info(self, qp: QP) -> Info:
    """Return info about a system state."""
    dp_c = self.box_plane.apply(qp, 1.)
    dp_c += self.box_heightMap.apply(qp, 1.)
    dp_c += self.capsule_plane.apply(qp, 1.)
    dp_c += self.capsule_capsule.apply(qp, 1.)

    dp_j = self.joint_revolute.apply(qp)
    dp_j += self.joint_universal.apply(qp)
    dp_j += self.joint_spherical.apply(qp)

    dp_a = P(jnp.zeros((self.num_bodies, 3)), jnp.zeros((self.num_bodies, 3)))

    info = Info(contact=dp_c, joint=dp_j, actuator=dp_a)
    return info

  @functools.partial(jax.jit, static_argnums=(0,))
  def step(self, qp: QP, act: jnp.ndarray) -> Tuple[QP, Info]:
    """Calculates a physics step for a system, returns next state and info."""

    def substep(carry, _):
      qp, info = carry
      dt = self.config.dt / self.config.substeps

      # apply kinetic step
      qp = integrators.kinetic(self.config, qp, dt, self.active_pos,
                               self.active_rot)

      # apply impulses arising from joints and actuators
      dp_j = self.joint_revolute.apply(qp)
      dp_j += self.joint_universal.apply(qp)
      dp_j += self.joint_spherical.apply(qp)
      dp_a = self.angle_1d.apply(qp, act)
      dp_a += self.angle_2d.apply(qp, act)
      dp_a += self.angle_3d.apply(qp, act)
      dp_a += self.torque_1d.apply(qp, act)
      dp_a += self.torque_2d.apply(qp, act)
      dp_a += self.torque_3d.apply(qp, act)
      qp = integrators.potential(self.config, qp, dp_j + dp_a, dt,
                                 self.active_pos, self.active_rot)

      # apply collision velocity updates
      dp_c = self.box_plane.apply(qp, dt)
      dp_c += self.box_heightMap.apply(qp, dt)
      dp_c += self.capsule_plane.apply(qp, dt)
      dp_c += self.capsule_capsule.apply(qp, dt)
      qp = integrators.potential_collision(self.config, qp, dp_c,
                                           self.active_pos, self.active_rot)

      info = Info(
          contact=info.contact + dp_c,
          joint=info.joint + dp_j,
          actuator=info.actuator + dp_a)

      return (qp, info), ()

    zero = P(jnp.zeros((self.num_bodies, 3)), jnp.zeros((self.num_bodies, 3)))
    info = Info(contact=zero, joint=zero, actuator=zero)
    (qp, info), _ = jax.lax.scan(substep, (qp, info), (), self.config.substeps)
    return qp, info



# Copyright 2021 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trains a humanoid to run in the +x direction."""

import brax
from brax.envs import env
from brax.physics import bodies
from brax.physics.base import take
import jax
import jax.numpy as jnp


class Humanoid(env.Env):
  """Trains a humanoid to run in the +x direction."""

  def __init__(self, **kwargs):
    super().__init__(_SYSTEM_CONFIG, **kwargs)
    body = bodies.Body.from_config(self.sys.config)
    body = take(body, body.idx[:-1])  # skip the floor body
    self.mass = body.mass.reshape(-1, 1)
    self.inertia = body.inertia

  def reset(self, rng: jnp.ndarray) -> env.State:
    """Resets the environment to an initial state."""
    qp = self.sys.default_qp()
    qp, info = self.sys.step(qp,
                             jax.random.uniform(rng, (self.action_size,)) * .5)
    obs = self._get_obs(qp, info, jnp.zeros(self.action_size))
    reward, done, zero = jnp.zeros(3)
    metrics = {
        'reward_linvel': zero,
        'reward_quadctrl': zero,
        'reward_alive': zero,
        'reward_impact': zero
    }
    return env.State(qp, obs, reward, done, metrics)

  def step(self, state: env.State, action: jnp.ndarray) -> env.State:
    """Run one timestep of the environment's dynamics."""
    qp, info = self.sys.step(state.qp, action)
    obs = self._get_obs(qp, info, action)

    pos_before = state.qp.pos[:-1]  # ignore floor at last index
    pos_after = qp.pos[:-1]  # ignore floor at last index
    com_before = jnp.sum(pos_before * self.mass, axis=0) / jnp.sum(self.mass)
    com_after = jnp.sum(pos_after * self.mass, axis=0) / jnp.sum(self.mass)
    lin_vel_cost = 1.25 * (com_after[0] - com_before[0]) / self.sys.config.dt
    quad_ctrl_cost = .01 * jnp.sum(jnp.square(action))
    # can ignore contact cost, see: https://github.com/openai/gym/issues/1541
    quad_impact_cost = 0.0
    alive_bonus = 5.0
    reward = lin_vel_cost - quad_ctrl_cost - quad_impact_cost + alive_bonus

    done = jnp.where(qp.pos[0, 2] < 0.65, x=1.0, y=0.0)
    done = jnp.where(qp.pos[0, 2] > 2.1, x=1.0, y=done)
    state.metrics.update(
        reward_linvel=lin_vel_cost,
        reward_quadctrl=quad_ctrl_cost,
        reward_alive=alive_bonus,
        reward_impact=quad_impact_cost)

    return state.replace(qp=qp, obs=obs, reward=reward, done=done)

  def _get_obs(self, qp: brax.QP, info: brax.Info,
               action: jnp.ndarray) -> jnp.ndarray:
    """Observe humanoid body position, velocities, and angles."""
    # some pre-processing to pull joint angles and velocities
    joint_1d_angle, joint_1d_vel = self.sys.joint_revolute.angle_vel(qp)
    joint_2d_angle, joint_2d_vel = self.sys.joint_universal.angle_vel(qp)
    joint_3d_angle, joint_3d_vel = self.sys.joint_spherical.angle_vel(qp)

    # qpos:
    # Z of the torso (1,)
    # orientation of the torso as quaternion (4,)
    # joint angles (8,)
    qpos = [
        qp.pos[0, 2:], qp.rot[0], joint_1d_angle[0], joint_2d_angle[0],
        joint_2d_angle[1], joint_3d_angle[0], joint_3d_angle[1],
        joint_3d_angle[2]
    ]

    # qvel:
    # velocity of the torso (3,)
    # angular velocity of the torso (3,)
    # joint angle velocities (8,)
    qvel = [
        qp.vel[0], qp.ang[0], joint_1d_vel[0], joint_2d_vel[0], joint_2d_vel[1],
        joint_3d_vel[0], joint_3d_vel[1], joint_3d_vel[2]
    ]

    # actuator forces
    torque_1d = take(action, self.sys.torque_1d.act_idx)
    torque_1d *= self.sys.torque_1d.strength
    torque_2d = take(action, self.sys.torque_2d.act_idx)
    torque_2d = torque_2d.reshape(torque_2d.shape[:-2] + (-1,))
    torque_2d *= jnp.repeat(self.sys.torque_2d.strength, 2)
    torque_3d = take(action, self.sys.torque_3d.act_idx)
    torque_3d = torque_3d.reshape(torque_3d.shape[:-2] + (-1,))
    torque_3d *= jnp.repeat(self.sys.torque_3d.strength, 3)
    qfrc_actuator = [torque_1d, torque_2d, torque_3d]

    # external contact forces:
    # delta velocity (3,), delta ang (3,) * num bodies in the system
    cfrc_ext = [info.contact.vel, info.contact.ang]
    # flatten bottom dimension
    cfrc_ext = [x.reshape(x.shape[:-2] + (-1,)) for x in cfrc_ext]

    # center of mass obs:
    body_pos = qp.pos[:-1]  # ignore floor at last index
    body_vel = qp.vel[:-1]  # ignore floor at last index

    com_vec = jnp.sum(body_pos * self.mass, axis=0) / jnp.sum(self.mass)
    com_vel = body_vel * self.mass / jnp.sum(self.mass)

    def v_outer(a):
      return jnp.outer(a, a)

    def v_cross(a, b):
      return jnp.cross(a, b)

    v_outer = jax.vmap(v_outer, in_axes=[0])
    v_cross = jax.vmap(v_cross, in_axes=[0, 0])

    disp_vec = body_pos - com_vec
    com_inert = self.inertia + self.mass.reshape(
        (11, 1, 1)) * ((jnp.linalg.norm(disp_vec, axis=1)**2.).reshape(
            (11, 1, 1)) * jnp.stack([jnp.eye(3)] * 11) - v_outer(disp_vec))

    cinert = [com_inert.reshape(-1)]

    square_disp = (1e-7 + (jnp.linalg.norm(disp_vec, axis=1)**2.)).reshape(
        (11, 1))
    com_angular_vel = (v_cross(disp_vec, body_vel) / square_disp)
    cvel = [com_vel.reshape(-1), com_angular_vel.reshape(-1)]

    return jnp.concatenate(qpos + qvel + cinert + cvel + qfrc_actuator +
                           cfrc_ext)


_SYSTEM_CONFIG = """
bodies {
  name: "torso"
  colliders {
    position {
    }
    rotation {
      x: -90.0
    }
    capsule {
      radius: 0.07
      length: 0.28
    }
  }
  colliders {
    position {
      z: 0.19
    }
    capsule {
      radius: 0.09
      length: 0.18
    }
  }
  colliders {
    position {
      x: -0.01
      z: -0.12
    }
    rotation {
      x: -90.0
    }
    capsule {
      radius: 0.06
      length: 0.24
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 8.907463
}
bodies {
  name: "lwaist"
  colliders {
    position {
    }
    rotation {
      x: -90.0
    }
    capsule {
      radius: 0.06
      length: 0.24
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 2.2619467
}
bodies {
  name: "pelvis"
  colliders {
    position {
      x: -0.02
    }
    rotation {
      x: -90.0
    }
    capsule {
      radius: 0.09
      length: 0.32
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 6.6161942
}
bodies {
  name: "right_thigh"
  colliders {
    position {
      y: 0.005
      z: -0.17
    }
    rotation {
      x: -178.31532
    }
    capsule {
      radius: 0.06
      length: 0.46014702
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 4.751751
}
bodies {
  name: "right_shin"
  colliders {
    position {
      z: -0.15
    }
    rotation {
      x: -180.0
    }
    capsule {
      radius: 0.049
      length: 0.398
      end: -1
    }
  }
  colliders {
    position {
      z: -0.35
    }
    capsule {
      radius: 0.075
      length: 0.15
      end: 1
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 4.5228419
}
bodies {
  name: "left_thigh"
  colliders {
    position {
      y: -0.005
      z: -0.17
    }
    rotation {
      x: 178.31532
    }
    capsule {
      radius: 0.06
      length: 0.46014702
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 4.751751
}
bodies {
  name: "left_shin"
  colliders {
    position {
      z: -0.15
    }
    rotation {
      x: -180.0
    }
    capsule {
      radius: 0.049
      length: 0.398
      end: -1
    }
  }
  colliders {
    position {
      z: -0.35
    }
    capsule {
      radius: 0.075
      length: 0.15
      end: 1
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 4.5228419
}
bodies {
  name: "right_upper_arm"
  colliders {
    position {
      x: 0.08
      y: -0.08
      z: -0.08
    }
    rotation {
      x: 135.0
      y: 35.26439
      z: -75.0
    }
    capsule {
      radius: 0.04
      length: 0.35712814
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.6610805
}
bodies {
  name: "right_lower_arm"
  colliders {
    position {
      x: 0.09
      y: 0.09
      z: 0.09
    }
    rotation {
      x: -45.0
      y: 35.26439
      z: 15.0
    }
    capsule {
      radius: 0.031
      length: 0.33912814
    }
  }
  colliders {
    position {
      x: 0.18
      y: 0.18
      z: 0.18
    }
    capsule {
      radius: 0.04
      length: 0.08
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.2295402
}
bodies {
  name: "left_upper_arm"
  colliders {
    position {
      x: 0.08
      y: 0.08
      z: -0.08
    }
    rotation {
      x: -135.0
      y: 35.26439
      z: 75.0
    }
    capsule {
      radius: 0.04
      length: 0.35712814
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.6610805
}
bodies {
  name: "left_lower_arm"
  colliders {
    position {
      x: 0.09
      y: -0.09
      z: 0.09
    }
    rotation {
      x: 45.0
      y: 35.26439
      z: -15.0
    }
    capsule {
      radius: 0.031
      length: 0.33912814
    }
  }
  colliders {
    position {
      x: 0.18
      y: -0.18
      z: 0.18
    }
    capsule {
      radius: 0.04
      length: 0.08
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.2295402
}
bodies {
  name: "floor"
  colliders {
    plane {
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen { all: true }
}
joints {
  name: "abdomen_z"
  stiffness: 15000.0
  parent: "torso"
  child: "lwaist"
  parent_offset {
    x: -0.01
    z: -0.195
  }
  child_offset {
    z: 0.065
  }
  rotation {
    y: -90.0
  }
  angular_damping: 20.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  angle_limit {
    min: -75.0
    max: 30.0
  }
}
joints {
  name: "abdomen_x"
  stiffness: 15000.0
  parent: "lwaist"
  child: "pelvis"
  parent_offset {
    z: -0.065
  }
  child_offset {
    z: 0.1
  }
  rotation {
    x: 90.0
  }
  angular_damping: 20.0
  angle_limit {
    min: -35.0
    max: 35.0
  }
}
joints {
  name: "right_hip_x"
  stiffness: 8000.0
  parent: "pelvis"
  child: "right_thigh"
  parent_offset {
    y: -0.1
    z: -0.04
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 20.0
  limit_strength: 2000.0
  angle_limit {
    min: -10.0
    max: 10.0
  }
  angle_limit {
    min: -30.0
    max: 70.0
  }
  angle_limit {
    min: -10.0
    max: 10.0
  }
}
joints {
  name: "right_knee"
  stiffness: 15000.0
  parent: "right_thigh"
  child: "right_shin"
  parent_offset {
    y: 0.01
    z: -0.383
  }
  child_offset {
    z: 0.02
  }
  rotation {
    z: -90.0
  }
  angular_damping: 20.0
  angle_limit {
    min: -160.0
    max: -2.0
  }
}
joints {
  name: "left_hip_x"
  stiffness: 8000.0
  parent: "pelvis"
  child: "left_thigh"
  parent_offset {
    y: 0.1
    z: -0.04
  }
  child_offset {
  }
  angular_damping: 20.0
  limit_strength: 2000.0
  angle_limit {
    min: -10.0
    max: 10.0
  }
  angle_limit {
    min: -30.0
    max: 70.0
  }
  angle_limit {
    min: -10.0
    max: 10.0
  }
}
joints {
  name: "left_knee"
  stiffness: 15000.0
  parent: "left_thigh"
  child: "left_shin"
  parent_offset {
    y: -0.01
    z: -0.383
  }
  child_offset {
    z: 0.02
  }
  rotation {
    z: -90.0
  }
  angular_damping: 20.0
  angle_limit {
    min: -160.0
    max: -2.0
  }
}
joints {
  name: "right_shoulder1"
  stiffness: 15000.0
  parent: "torso"
  child: "right_upper_arm"
  parent_offset {
    y: -0.17
    z: 0.06
  }
  child_offset {
  }
  rotation {
    x: 135.0
    y: 35.26439
  }
  angular_damping: 20.0
  angle_limit {
    min: -85.0
    max: 60.0
  }
  angle_limit {
    min: -85.0
    max: 60.0
  }
}
joints {
  name: "right_elbow"
  stiffness: 15000.0
  parent: "right_upper_arm"
  child: "right_lower_arm"
  parent_offset {
    x: 0.18
    y: -0.18
    z: -0.18
  }
  child_offset {
  }
  rotation {
    x: 135.0
    z: 90.0
  }
  angular_damping: 20.0
  angle_limit {
    min: -90.0
    max: 50.0
  }
}
joints {
  name: "left_shoulder1"
  stiffness: 15000.0
  parent: "torso"
  child: "left_upper_arm"
  parent_offset {
    y: 0.17
    z: 0.06
  }
  child_offset {
  }
  rotation {
    x: 45.0
    y: -35.26439
  }
  angular_damping: 20.0
  angle_limit {
    min: -60.0
    max: 85.0
  }
  angle_limit {
    min: -60.0
    max: 85.0
  }
}
joints {
  name: "left_elbow"
  stiffness: 15000.0
  parent: "left_upper_arm"
  child: "left_lower_arm"
  parent_offset {
    x: 0.18
    y: 0.18
    z: -0.18
  }
  child_offset {
  }
  rotation {
    x: 45.0
    z: -90.0
  }
  angular_damping: 20.0
  angle_limit {
    min: -90.0
    max: 50.0
  }
}
actuators {
  name: "abdomen_z"
  joint: "abdomen_z"
  strength: 300.0
  torque {
  }
}
actuators {
  name: "abdomen_x"
  joint: "abdomen_x"
  strength: 300.0
  torque {
  }
}
actuators {
  name: "right_hip_x"
  joint: "right_hip_x"
  strength: 300.0
  torque {
  }
}
actuators {
  name: "right_knee"
  joint: "right_knee"
  strength: 300.0
  torque {
  }
}
actuators {
  name: "left_hip_x"
  joint: "left_hip_x"
  strength: 300.0
  torque {
  }
}
actuators {
  name: "left_knee"
  joint: "left_knee"
  strength: 300.0
  torque {
  }
}
actuators {
  name: "right_shoulder1"
  joint: "right_shoulder1"
  strength: 75.0
  torque {
  }
}
actuators {
  name: "right_elbow"
  joint: "right_elbow"
  strength: 75.0
  torque {
  }
}
actuators {
  name: "left_shoulder1"
  joint: "left_shoulder1"
  strength: 75.0
  torque {
  }
}
actuators {
  name: "left_elbow"
  joint: "left_elbow"
  strength: 75.0
  torque {
  }
}
collide_include {
  first: "floor"
  second: "left_shin"
}
collide_include {
  first: "floor"
  second: "right_shin"
}
defaults {
  angles {
    name: "left_knee"
    angle { x: -25. y: 0 z: 0 }
  }
  angles {
    name: "right_knee"
    angle { x: -25. y: 0 z: 0 }
  }
}
friction: 1.0
gravity {
  z: -9.81
}
angular_damping: -0.05
baumgarte_erp: 0.1
dt: 0.015
substeps: 8
"""
